#
# This file is part of TransportMaps.
#
# TransportMaps is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# TransportMaps is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with TransportMaps.  If not, see <http://www.gnu.org/licenses/>.
#
# Transport Maps Library
# Copyright (C) 2015-2018 Massachusetts Institute of Technology
# Uncertainty Quantification group
# Department of Aeronautics and Astronautics
#
# Authors: Transport Map Team
# Website: transportmaps.mit.edu
# Support: transportmaps.mit.edu/qa/
#

import numpy as np

from TransportMaps.Misc import counted, cached, cached_tuple, get_sub_cache
from TransportMaps.Functionals.FunctionBase import *
from TransportMaps.Maps.MapBase import \
    Map, ConstantMap, LinearMap, ConditionallyLinearMap
from TransportMaps.Distributions.DistributionBase import \
    ConditionalDistribution
from TransportMaps.Distributions.FrozenDistributions import \
    GaussianDistribution
from TransportMaps.Distributions.ConditionalDistributions import \
    ConditionallyGaussianDistribution

__all__ = ['LogLikelihood',
           'AdditiveLogLikelihood',
           'AdditiveLinearGaussianLogLikelihood',
           'AdditiveConditionallyLinearGaussianLogLikelihood',
           'IndependentLogLikelihood']

class LogLikelihood(Function):
    r""" Abstract class for log-likelihood :math:`\log \pi({\bf y} \vert {\bf x})`

    Note that :math:`\log\pi:\mathbb{R}^d \rightarrow \mathbb{R}`
    is considered a function of :math:`{\bf x}`, while the
    data :math:`{\bf y}` is fixed.
    
    Args:
      y (:class:`ndarray<numpy.ndarray>`): data
      dim (int): input dimension $d$
    """
    def __init__(self, y, dim):
        super(LogLikelihood, self).__init__(dim)
        self._y = y

    @property
    def y(self):
        return self._y

    @cached()
    @counted
    def evaluate(self, x, *args, **kwargs):
        r""" [Abstract] Evaluate :math:`\log\pi({\bf y} \vert {\bf x})`.

        Args:
          x (:class:`ndarray<numpy.ndarray>` [:math:`m,d`]): evaluation points

        Returns:
          (:class:`ndarray<numpy.ndarray>` [:math:`m`]) -- function evaluations
        """
        raise NotImplementedError("To be implemented in sub-classes")

    @cached()
    @counted
    def grad_x(self, x, *args, **kwargs):
        r""" [Abstract] Evaluate :math:`\nabla_{\bf x}\log\pi({\bf y} \vert {\bf x})`.

        Args:
          x (:class:`ndarray<numpy.ndarray>` [:math:`m,d`]): evaluation points

        Returns:
          (:class:`ndarray<numpy.ndarray>` [:math:`m,d`]) -- gradient evaluations
        """
        raise NotImplementedError("To be implemented in sub-classes")

    @cached_tuple(['evaluate', 'grad_x'])
    @counted 
    def tuple_grad_x(self, x, cache=None, **kwargs):
        r""" Evaluate :math:`\left(\log\pi({\bf y} \vert {\bf x}),\nabla_{\bf x}\log\pi({\bf y} \vert {\bf x})\right)`.

        Args:
          x (:class:`ndarray<numpy.ndarray>` [:math:`m,d`]): evaluation points

        Returns:
          (:class:`tuple`) --
            :math:`\left(\log\pi({\bf y} \vert {\bf x}),\nabla_{\bf x}\log\pi({\bf y} \vert {\bf x})\right)`
        """
        return ( self.evaluate(x, cache=cache, **kwargs),
                 self.grad_x(x, cache=cache, **kwargs) )

    @cached(caching=False)
    @counted
    def hess_x(self, x, *args, **kwargs):
        r""" [Abstract] Evaluate :math:`\nabla^2_{\bf x}\log\pi({\bf y} \vert {\bf x})`.

        Args:
          x (:class:`ndarray<numpy.ndarray>` [:math:`m,d`]): evaluation points

        Returns:
          (:class:`ndarray<numpy.ndarray>` [:math:`m,d,d`]) -- Hessian evaluations
        """
        raise NotImplementedError("To be implemented in sub-classes")

    @cached(caching=False)
    @counted
    def action_hess_x(self, x, dx, *args, **kwargs):
        r""" [Abstract] Evaluate :math:`\langle \nabla^2_{\bf x}\log\pi({\bf y} \vert {\bf x}),\delta{\bf x}\rangle`.

        Args:
          x (:class:`ndarray<numpy.ndarray>` [:math:`m,d`]): evaluation points
          dx (:class:`ndarray<numpy.ndarray>` [:math:`m,d`]): direction
            on which to evaluate the Hessian

        Returns:
          (:class:`ndarray<numpy.ndarray>` [:math:`m,d`]) -- Hessian evaluations
        """
        raise NotImplementedError("To be implemented in sub-classes")

class AdditiveLogLikelihood(LogLikelihood):
    r""" Log-likelihood :math:`\log \pi({\bf y} \vert {\bf x})=\log\pi({\bf y} - T({\bf x}))`

    Args:
      y (:class:`ndarray<numpy.ndarray>` :math:`[d_y]`): data
      pi (:class:`Distribution<TransportMaps.Distributions.Distribution>`):
        distribution :math:`\nu_\pi`
      T (:class:`Map<TransportMaps.Maps.Map>`): map :math:`T:\mathbb{R}^d\rightarrow\mathbb{R}^{d_y}`
    """
    def __init__(self, y, pi, T):
        if len(y) != pi.dim:
            raise ValueError("The dimension of the data must match the " + \
                             "dimension of the distribution pi")
        if len(y) != T.dim_out:
            raise ValueError("The dimension of the data must match the " + \
                             "dimension of the output of T")
        super(AdditiveLogLikelihood, self).__init__(y, T.dim_in)
        self.pi = pi
        self.T = T
        if issubclass(type(pi), ConditionalDistribution):
            self._isPiCond = True
        else:
            self._isPiCond = False

    def get_ncalls_tree(self, indent=""):
        out = super(AdditiveLogLikelihood, self).get_ncalls_tree(indent)
        out += self.pi.get_ncalls_tree(indent + '  ')
        out += self.T.get_ncalls_tree(indent + '  ')
        return out

    def get_nevals_tree(self, indent=""):
        out = super(AdditiveLogLikelihood, self).get_nevals_tree(indent)
        out += self.pi.get_nevals_tree(indent + '  ')
        out += self.T.get_nevals_tree(indent + '  ')
        return out

    def get_teval_tree(self, indent=""):
        out = super(AdditiveLogLikelihood, self).get_teval_tree(indent)
        out += self.pi.get_teval_tree(indent + '  ')
        out += self.T.get_teval_tree(indent + '  ')
        return out

    def update_ncalls_tree(self, obj):
        super(AdditiveLogLikelihood, self).update_ncalls_tree(obj)
        self.pi.update_ncalls_tree(obj.pi)
        self.T.update_ncalls_tree(obj.T)

    def update_nevals_tree(self, obj):
        super(AdditiveLogLikelihood, self).update_nevals_tree(obj)
        self.pi.update_nevals_tree(obj.pi)
        self.T.update_nevals_tree(obj.T)

    def update_teval_tree(self, obj):
        super(AdditiveLogLikelihood, self).update_teval_tree(obj)
        self.pi.update_teval_tree(obj.pi)
        self.T.update_teval_tree(obj.T)

    def reset_counters(self):
        super(AdditiveLogLikelihood, self).reset_counters()
        self.pi.reset_counters()
        self.T.reset_counters()
        
    @property
    def isPiCond(self):
        return self._isPiCond

    @cached([('pi',None),('T',None)])
    @counted
    def evaluate(self, x, idxs_slice=slice(None,None,None), cache=None, **kwargs):
        r""" Evaluate :math:`\log\pi({\bf y} \vert {\bf x})`.

        Args:
          x (:class:`ndarray<numpy.ndarray>` [:math:`m,d`]): evaluation points
          cache (dict): cache

        Returns:
          (:class:`ndarray<numpy.ndarray>` [:math:`m`]) -- function evaluations
        """
        pi_cache, T_cache = get_sub_cache(cache, ('pi',None), ('T',None))
        frw = self.T.evaluate(x, idxs_slice=idxs_slice, cache=T_cache)
        return self.pi.log_pdf(
            self.y - frw, idxs_slice=idxs_slice, cache=pi_cache )

    @cached([('pi',None),('T',None)])
    @counted
    def grad_x(self, x, idxs_slice=slice(None,None,None), cache=None, **kwargs):
        r""" Evaluate :math:`\nabla_{\bf x}\log\pi({\bf y} \vert {\bf x})`.

        Args:
          x (:class:`ndarray<numpy.ndarray>` [:math:`m,d`]): evaluation points
          cache (dict): cache

        Returns:
          (:class:`ndarray<numpy.ndarray>` [:math:`m,d`]) -- gradient evaluations

        .. todo:: caching is not implemented
        """
        pi_cache, T_cache = get_sub_cache(cache, ('pi',None), ('T',None))
        frw = self.T.evaluate(x, cache=T_cache)
        gx_frw = self.T.grad_x(x, idxs_slice=idxs_slice, cache=T_cache)
        gx_lpdf = self.pi.grad_x_log_pdf(
            self.y - frw, idxs_slice=idxs_slice, cache=pi_cache )
        out = - np.einsum('...i,...ij->...j',gx_lpdf, gx_frw)
        return out

    @cached_tuple(['evaluate','grad_x'], [('pi',None),('T',None)])
    @counted
    def tuple_grad_x(self, x, idxs_slice=slice(None,None,None), cache=None, **kwargs):
        r""" Evaluate :math:`\log\pi({\bf y} \vert {\bf x}), \nabla_{\bf x}\log\pi({\bf y} \vert {\bf x})`.

        Args:
          x (:class:`ndarray<numpy.ndarray>` [:math:`m,d`]): evaluation points
          cache (dict): cache

        Returns:
          (:class:`tuple`) -- function and gradient evaluations

        .. todo:: caching is not implemented
        """
        pi_cache, T_cache = get_sub_cache(cache, ('pi',None), ('T',None))
        frw, gx_frw = self.T.tuple_grad_x(x, idxs_slice=idxs_slice, cache=T_cache)
        lpdf, gx_lpdf = self.pi.tuple_grad_x_log_pdf(
            self.y - frw, idxs_slice=idxs_slice, cache=pi_cache )
        gx = - np.einsum('...i,...ij->...j',gx_lpdf, gx_frw)
        return lpdf, gx

    @cached([('pi',None),('T',None)],False)
    @counted
    def hess_x(self, x, idxs_slice=slice(None,None,None), cache=None, **kwargs):
        r""" Evaluate :math:`\nabla^2_{\bf x}\log\pi({\bf y} \vert {\bf x})`.

        Args:
          x (:class:`ndarray<numpy.ndarray>` [:math:`m,d`]): evaluation points
          cache (dict): cache

        Returns:
          (:class:`ndarray<numpy.ndarray>` [:math:`m,d,d`]) -- Hessian evaluations

        .. todo:: caching is not implemented
        """
        pi_cache, T_cache = get_sub_cache(cache, ('pi',None), ('T',None))
        frw = self.T.evaluate(x, idxs_slice=idxs_slice, cache=T_cache) # m x d_y
        gx_frw = self.T.grad_x(x, idxs_slice=idxs_slice, cache=T_cache) # m x d_y x d_x
        hx_frw = self.T.hess_x(x, idxs_slice=idxs_slice, cache=T_cache) # m x d_y x d_x x d_x
        gx_lpdf = self.pi.grad_x_log_pdf(
            self.y - frw, idxs_slice=idxs_slice, cache=pi_cache ) # m x d_y
        hx_lpdf = self.pi.hess_x_log_pdf(
            self.y - frw, idxs_slice=idxs_slice, cache=pi_cache ) # m x d_y x d_y
        out = np.einsum('...ij,...ik,...kl->...jl', gx_frw, hx_lpdf, gx_frw)
        out -= np.einsum('...i,...ijk->...jk', gx_lpdf, hx_frw) # m x d_x x d_x
        return out

    @cached([('pi',None),('T',None)],False)
    @counted
    def action_hess_x(self, x, dx, idxs_slice=slice(None,None,None), cache=None, **kwargs):
        r""" Evaluate :math:`\langle\nabla^2_{\bf x}\log\pi({\bf y} \vert {\bf x}),\delta{\bf x}\rangle`.

        Args:
          x (:class:`ndarray<numpy.ndarray>` [:math:`m,d`]): evaluation points
          dx (:class:`ndarray<numpy.ndarray>` [:math:`m,d`]): direction
            on which to evaluate the Hessian
          cache (dict): cache

        Returns:
          (:class:`ndarray<numpy.ndarray>` [:math:`m,d,`]) -- action of Hessian evaluations

        .. todo:: caching is not implemented
        """
        pi_cache, T_cache = get_sub_cache(cache, ('pi',None), ('T',None))
        
        frw = self.T.evaluate(x, idxs_slice=idxs_slice, cache=T_cache) # m x d_y
        gx_frw = self.T.grad_x(x, idxs_slice=idxs_slice, cache=T_cache) # m x d_y x d_x
        gx_lpdf = self.pi.grad_x_log_pdf(
            self.y - frw, idxs_slice=idxs_slice, cache=pi_cache ) # m x d_y

        A = np.einsum('...ij,...j->...i', -gx_frw, dx) # m x d_y
        A = self.pi.action_hess_x_log_pdf(
            self.y - frw, A, idxs_slice=idxs_slice, cache=pi_cache) # m x d_y
        A = np.einsum('...ij,...i->...j', -gx_frw, A) # m x d_x

        B = -self.T.action_hess_x(
            x, dx, idxs_slice=idxs_slice, cache=T_cache) # m x d_y x d_x
        B = np.einsum('...i,...ij->...j', gx_lpdf, B) # m x d_x

        return A + B

class AdditiveLinearGaussianLogLikelihood(AdditiveLogLikelihood):
    r""" Define the log-likelihood for the additive linear Gaussian model

    The model is

    .. math::

       {\bf y} = {\bf c} + {\bf T}{\bf x} + \varepsilon \;, \quad
       \varepsilon \sim \mathcal{N}(\mu, \Sigma)

    where :math:`T \in \mathbb{R}^{d_y \times d_x}`, :math:`\mu \in \mathbb{R}^{d_y}`
    and :math:`\Sigma \in \mathbb{R}^{d_y \times d_y}` is symmetric positve
    definite

    Args:
      y (:class:`ndarray<numpy.ndarray>` [:math:`d_y`]): data
      c (:class:`ndarray<numpy.ndarray>` [:math:`d_y`] or :class:`Map<TransportMaps.Maps.Map>`): system constant or parametric map returning the constant
      T (:class:`ndarray<numpy.ndarray>` [:math:`d_y,d_x`] or :class:`Map<TransportMaps.Maps.Map>`): system matrix or parametric map returning the system matrix
      mu (:class:`ndarray<numpy.ndarray>` [:math:`d_y`] or :class:`Map<TransportMaps.Maps.Map>`): noise mean or parametric map returning the mean
      sigma (:class:`ndarray<numpy.ndarray>` [:math:`d_y,d_y`] or :class:`Map<TransportMaps.Maps.Map>`):
        noise covariance or parametric map returning the covariance
      precision (:class:`ndarray<numpy.ndarray>` [:math:`d_y,d_y`] or :class:`Map<TransportMaps.Maps.Map>`):
        noise precision matrix or parametric map returning the precision matrix
    """
    def __init__(self, y, c, T, mu, sigma=None, precision=None):
        # INIT MAP AND DISTRIBUTION
        linmap = LinearMap(c, T)
        pi = GaussianDistribution(mu, sigma=sigma, precision=precision)
        super(AdditiveLinearGaussianLogLikelihood, self).__init__(y, pi, linmap)

class AdditiveConditionallyLinearGaussianLogLikelihood(AdditiveLogLikelihood):
    r""" Define the log-likelihood for the additive linear Gaussian model

    The model is

    .. math::

       {\bf y} = {\bf c}(\theta) + {\bf T}(\theta){\bf x} + \varepsilon \;, \quad
       \varepsilon \sim \mathcal{N}(\mu(\theta), \Sigma(\theta))

    where :math:`T \in \mathbb{R}^{d_y \times d_x}`, :math:`\mu \in \mathbb{R}^{d_y}`
    and :math:`\Sigma \in \mathbb{R}^{d_y \times d_y}` is symmetric positve
    definite

    Args:
      y (:class:`ndarray<numpy.ndarray>` [:math:`d_y`]): data
      c (:class:`ndarray<numpy.ndarray>` [:math:`d_y`] or :class:`Map<TransportMaps.Maps.Map>`): system constant or parametric map returning the constant
      T (:class:`ndarray<numpy.ndarray>` [:math:`d_y,d_x`] or :class:`Map<TransportMaps.Maps.Map>`): system matrix or parametric map returning the system matrix
      mu (:class:`ndarray<numpy.ndarray>` [:math:`d_y`] or :class:`Map<TransportMaps.Maps.Map>`): noise mean or parametric map returning the mean
      sigma (:class:`ndarray<numpy.ndarray>` [:math:`d_y,d_y`] or :class:`Map<TransportMaps.Maps.Map>`):
        noise covariance or parametric map returning the covariance
      precision (:class:`ndarray<numpy.ndarray>` [:math:`d_y,d_y`] or :class:`Map<TransportMaps.Maps.Map>`):
        noise precision matrix or parametric map returning the precision matrix
      active_vars_system (:class:`list` of :class:`int`): active variables
        identifying the parameters for for :math:`c(\theta), T(\theta)`.
      active_vars_distribution (:class:`list` of :class:`int`): active variables
        identifying the parameters for for :math:`\mu(\theta), \Sigma(\theta)`.
      coeffs (:class:`ndarray<numpy.ndarray>`): initialization coefficients
    """
    def __init__(self, y, c, T, mu, sigma=None, precision=None,
                 active_vars_system=[], active_vars_distribution=[],
                 coeffs=None):
        # SYSTEM
        if c.dim_in != T.dim_in:
            raise ValueError("Input dimension c and T don't match")
        # DISTRIBUTION
        if isinstance(mu, np.ndarray) and (
                (sigma is not None and isinstance(sigma, np.ndarray)) or
                (precision is not None and isinstance(precision, np.ndarray)) ):
            pi = GaussianDistribution(mu, sigma=sigma, precision=precision)
        else:
            if mu.dim_in != c.dim_in:
                raise ValueError("Input dimension c and mu don't match")
            if sigma is not None and mu.dim_in != sigma.dim_in:
                raise ValueError("Input dimension mu and sigma don't match")
            if precision is not None and mu.dim_in != precision.dim_in:
                raise ValueError("Input dimension mu and precision don't match")
            pi = ConditionallyGaussianDistribution(mu, sigma=sigma, precision=precision)
        # SETUP AUXILIARY VARIABLES
        self._n_coeffs = c.dim_in
        # INIT MAP AND DISTRIBUTION
        linmap = ConditionallyLinearMap(c, T)
        super(AdditiveConditionallyLinearGaussianLogLikelihood, self).__init__(y, pi, linmap)
        self.coeffs = coeffs

    @property
    def n_coeffs(self):
        return self._n_coeffs
        
    @property
    def coeffs(self):
        return self._coeffs

    @coeffs.setter
    def coeffs(self, coeffs):
        if coeffs is not None:
            self.T.coeffs = coeffs
            if self.isPiCond:
                self.pi.coeffs = coeffs
            self._coeffs = coeffs

            
class IndependentLogLikelihood(Function):
    r""" Handling likelihoods in the form :math:`\pi({\bf y}\vert{\bf x}) = \prod_{i=1}^{n}\pi_i({\bf y}_i\vert{\bf x}_i)`

    Args:
      factors (:class:`list` of :class:`tuple`): each tuple contains a
        log-likelihood (:class:`LogLikelihood`) and the sub-set of variables
        of :math:`{\bf x}` on which it acts.

    Example
    -------
    Let :math:`\pi(y_0,y_2\vert x_0,x_1,x_2)= \pi_0(y_0\vert x_0) \pi_2(y_2,x_2)`.

    >>> factors = [(ll0, [0]),
    >>>            (ll2, [2])]
    >>> ll = IndependentLogLikelihood(factors)
    
    """
    def __init__(self, factors):
        self.factors = []
        self.input_dim = set()
        for i, (ll, xidxs) in enumerate(factors):
            # Check right number of inputs
            if ll is not None and len(set(list(xidxs))) != ll.dim:
                raise TypeError("The dimension of the %d " % i + \
                                "log-likelihood is not consistent with " + \
                                "the number of variables.")
            self.factors.append( (ll, xidxs) )
            self.input_dim |= set(xidxs)
        dim = 0 if len(self.input_dim) == 0 else max(self.input_dim)+1
        super(IndependentLogLikelihood, self).__init__(dim)

    def get_ncalls_tree(self, indent=""):
        out = super(IndependentLogLikelihood, self).get_ncalls_tree(indent)
        for ll,_ in self.factors:
            out += ll.get_ncalls_tree(indent + '  ')
        return out

    def get_nevals_tree(self, indent=""):
        out = super(IndependentLogLikelihood, self).get_nevals_tree(indent)
        for ll,_ in self.factors:
            out += ll.get_nevals_tree(indent + '  ')
        return out

    def get_teval_tree(self, indent=""):
        out = super(IndependentLogLikelihood, self).get_teval_tree(indent)
        for ll,_ in self.factors:
            out += ll.get_teval_tree(indent + '  ')
        return out

    def update_ncalls_tree(self, obj):
        super(IndependentLogLikelihood, self).update_ncalls_tree(obj)
        for (ll,_),(obj_ll,_) in zip(self.factors,obj.factors):
            ll.update_ncalls_tree(obj_ll)

    def update_nevals_tree(self, obj):
        super(IndependentLogLikelihood, self).update_nevals_tree(obj)
        for (ll,_),(obj_ll,_) in zip(self.factors,obj.factors):
            ll.update_nevals_tree(obj_ll)

    def update_teval_tree(self, obj):
        super(IndependentLogLikelihood, self).update_teval_tree(obj)
        for (ll,_),(obj_ll,_) in zip(self.factors,obj.factors):
            ll.update_teval_tree(obj_ll)

    def reset_counters(self):
        super(IndependentLogLikelihood, self).reset_counters()
        for ll,_ in self.factors:
            ll.reset_counters()
            
    @property
    def y(self):
        return [ ll.y for ll,_ in self.factors ]

    @property
    def n_factors(self):
        return len(self.factors)

    def append(self, factor):
        r""" Add a new factor to the likelihood

        Args:
          factors (:class:`tuple`): tuple containing a
            log-likelihood (:class:`LogLikelihood`) and the sub-set of variables
            of :math:`{\bf x}` on which it acts.

        Example
        -------
        Let :math:`\pi(y_0,y_2\vert x_0,x_1,x_2)= \pi_0(y_0\vert x_0) \pi_2(y_2,x_2)` and
        let's add the factor :math:`\pi_1(y_1\vert x_1)`, obtaining:

        .. math::

           \pi(y_0,y_1,y_2\vert x_0,x_1,x_2)= \pi_0(y_0\vert x_0) \pi_1(y_1\vert x_1) \pi_2(y_2,x_2)

        >>> factor = (ll1, [1])
        >>> ll.append(factor)
        
        """
        ll, xidxs = factor
        if ll is not None and len(set(xidxs)) != ll.dim:
            raise TypeError("The dimension of the " + \
                            "log-likelihood is not consistent with " + \
                            "the number of variables.")
        self.factors.append( (ll, xidxs) )
        self.input_dim |= set(xidxs)
        self.dim = max(self.input_dim)+1

    @cached([("ll_list","n_factors")])
    @counted
    def evaluate(self, x, idxs_slice=slice(None,None,None), cache=None, **kwargs):
        r""" Evaluate :math:`\log\pi({\bf y} \vert {\bf x})`.

        Args:
          x (:class:`ndarray<numpy.ndarray>` [:math:`m,d`]): evaluation points
          cache (dict): cache

        Returns:
          (:class:`ndarray<numpy.ndarray>` [:math:`m`]) -- function evaluations
        """
        if x.shape[1] != self.dim:
            raise ValueError("The dimension of the input does not match the dimension " + \
                             "of the log-likelihood")
        ll_list_cache = get_sub_cache(cache, ("ll_list",self.n_factors))
        out = np.zeros(x.shape[0])
        for (ll, xidxs), ll_cache in zip(self.factors, ll_list_cache):
            if ll is not None:
                out += ll.evaluate(
                    x[:,xidxs], idxs_slice=idxs_slice, cache=ll_cache, **kwargs)
        return out

    @cached([("ll_list","n_factors")])
    @counted
    def grad_x(self, x, idxs_slice=slice(None,None,None), cache=None, **kwargs):
        r""" Evaluate :math:`\nabla_{\bf x}\log\pi({\bf y} \vert {\bf x})`.

        Args:
          x (:class:`ndarray<numpy.ndarray>` [:math:`m,d`]): evaluation points
          cache (dict): cache

        Returns:
          (:class:`ndarray<numpy.ndarray>` [:math:`m,d`]) -- gradient evaluations
        """
        if x.shape[1] != self.dim:
            raise ValueError("The dimension of the input does not match the dimension " + \
                             "of the log-likelihood")
        ll_list_cache = get_sub_cache(cache, ("ll_list",self.n_factors))
        out = np.zeros(x.shape)
        for (ll, xidxs), ll_cache in zip(self.factors, ll_list_cache):
            if ll is not None:
                out[:,xidxs] += ll.grad_x(
                    x[:,xidxs], idxs_slice=idxs_slice, cache=ll_cache, **kwargs)
        return out

    @cached_tuple(['evaluate', 'grad_x'],[("ll_list","n_factors")])
    @counted
    def tuple_grad_x(self, x, idxs_slice=slice(None,None,None), cache=None, **kwargs):
        r""" Evaluate :math:`\log\pi({\bf y} \vert {\bf x}), \nabla_{\bf x}\log\pi({\bf y} \vert {\bf x})`.

        Args:
          x (:class:`ndarray<numpy.ndarray>` [:math:`m,d`]): evaluation points
          cache (dict): cache

        Returns:
          (:class:`tuple`) -- function and gradient evaluations
        """
        if x.shape[1] != self.dim:
            raise ValueError("The dimension of the input does not match the dimension " + \
                             "of the log-likelihood")
        ll_list_cache = get_sub_cache(cache, ("ll_list",self.n_factors))
        fx_out = np.zeros(x.shape[0])
        gx_out = np.zeros(x.shape)
        for (ll, xidxs), ll_cache in zip(self.factors, ll_list_cache):
            if ll is not None:
                fx, gfx = ll.tuple_grad_x(
                    x[:,xidxs], idxs_slice=idxs_slice, cache=ll_cache, **kwargs)
                fx_out += fx
                gx_out[:,xidxs] += gfx
        return fx_out, gx_out
        
    @cached([("ll_list","n_factors")],False)
    @counted
    def hess_x(self, x, idxs_slice=slice(None,None,None), cache=None, **kwargs):
        r""" Evaluate :math:`\nabla^2_{\bf x}\log\pi({\bf y} \vert {\bf x})`.

        Args:
          x (:class:`ndarray<numpy.ndarray>` [:math:`m,d`]): evaluation points
          cache (dict): cache

        Returns:
          (:class:`ndarray<numpy.ndarray>` [:math:`m,d,d`]) -- Hessian evaluations
        """
        if x.shape[1] != self.dim:
            raise ValueError("The dimension of the input does not match the dimension " + \
                             "of the log-likelihood")
        ll_list_cache = get_sub_cache(cache, ("ll_list",self.n_factors))
        m = x.shape[0]
        out = np.zeros( (m, self.dim, self.dim) )
        for (ll, xidxs), ll_cache in zip(self.factors, ll_list_cache):
            if ll is not None:
                out[np.ix_(range(m),xidxs,xidxs)] += \
                    ll.hess_x(
                        x[:,xidxs], idxs_slice=idxs_slice, cache=ll_cache, **kwargs)
        return out

    @cached([("ll_list","n_factors")],False)
    @counted
    def action_hess_x(self, x, dx, idxs_slice=slice(None,None,None), cache=None, **kwargs):
        r""" Evaluate :math:`\langle\nabla^2_{\bf x}\log\pi({\bf y} \vert {\bf x}),\delta{\bf x}\rangle`.

        Args:
          x (:class:`ndarray<numpy.ndarray>` [:math:`m,d`]): evaluation points
          dx (:class:`ndarray<numpy.ndarray>` [:math:`m,d`]): direction
            on which to evaluate the Hessian
          cache (dict): cache

        Returns:
          (:class:`ndarray<numpy.ndarray>` [:math:`m,d`]) -- Hessian evaluations
        """
        if x.shape[1] != self.dim:
            raise ValueError("The dimension of the input does not match the dimension " + \
                             "of the log-likelihood")
        ll_list_cache = get_sub_cache(cache, ("ll_list",self.n_factors))
        m = x.shape[0]
        out = np.zeros( (m, self.dim) )
        for (ll, xidxs), ll_cache in zip(self.factors, ll_list_cache):
            if ll is not None:
                out[np.ix_(range(m),xidxs)] += \
                    ll.action_hess_x(
                        x[:,xidxs], dx[:,xidxs],
                        idxs_slice=idxs_slice, cache=ll_cache, **kwargs)
        return out


from . import SparsityIdentificationNonGaussian
from .SparsityIdentificationNonGaussian import *
from . import NodeOrdering
from .NodeOrdering import *
from . import GeneralizedPrecision
from .GeneralizedPrecision import *

__all__ = []
__all__ += SparsityIdentificationNonGaussian.__all__
__all__ += NodeOrdering.__all__
__all__ += GeneralizedPrecision.__all__




#
# This file is part of TransportMaps.
#
# TransportMaps is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# TransportMaps is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with TransportMaps.  If not, see <http://www.gnu.org/licenses/>.
#
# Transport Maps Library
# Copyright (C) 2015-2018 Massachusetts Institute of Technology
# Uncertainty Quantification group
# Department of Aeronautics and Astronautics
#
# Author: Transport Map Team
# Website: transportmaps.mit.edu
# Support: transportmaps.mit.edu/qa/
#

import numpy as np

from TransportMaps import mpi_map
from TransportMaps.Samplers.SamplerBase import *

__all__ = ['ImportanceSampler', 'RejectionSampler']

class ImportanceSampler(Sampler):
    r""" Importance sampler of distribution ``d``, with biasing distribution ``d_bias``

    Args:
      d (Distributions.Distribution): distribution to sample from
      d_bias (Distributions.Distribution): biasing distribution
    """
    def __init__(self, d, d_bias):
        if d.dim != d_bias.dim:
            raise ValueError("Dimension of the densities ``d`` and ``d_bias`` must " + \
                             "be the same")
        super(ImportanceSampler, self).__init__(d)
        self.bias_distribution = d_bias

    def rvs(self, m, mpi_pool_tuple=(None, None)):
        r""" Generate :math:`m` samples and importance weights from the distribution

        Args:
          m (int): number of samples to generate

        Returns:
          (:class:`tuple` (:class:`ndarray<numpy.ndarray>` [:math:`m,d`], :class:`ndarray<numpy.ndarray>` [:math:`m`])) -- list of points and weights
        """
        samps = self.bias_distribution.rvs(m, mpi_pool=mpi_pool_tuple[1])
        scatter_tuple = (['x'], [samps])
        num = mpi_map('pdf', obj=self.distribution, scatter_tuple=scatter_tuple,
                      mpi_pool=mpi_pool_tuple[0])
        den = mpi_map('pdf', obj=self.bias_distribution, scatter_tuple=scatter_tuple,
                      mpi_pool=mpi_pool_tuple[1])
        weights = num/den
        weights /= np.sum(weights)
        return (samps, weights)

class RejectionSampler(Sampler):
    r""" Rejection sampler of distribution ``d``, with biasing distribution ``d_bias``

    Args:
      d (Distributions.Distribution): distribution to sample from
      d_bias (Distributions.Distribution): biasing distribution
    """
    def __init__(self, d, d_bias):
        if d.dim != d_bias.dim:
            raise ValueError("Dimension of the densities ``d`` and ``d_bias`` must " + \
                             "be the same")
        super(RejectionSampler, self).__init__(d)
        self.bias_distribution = d_bias

    def rvs(self, m, *args, **kwargs):
        r""" Generate :math:`m` samples and importance weights from the distribution

        Args:
          m (int): number of samples to generate

        Returns:
          (:class:`tuple` (:class:`ndarray<numpy.ndarray>` [:math:`m,d`], :class:`ndarray<numpy.ndarray>` [:math:`m`])) -- list of points and weights
        """
        raise NotImplementedError("bias_distribution needs to dominate...")
        samps = np.zeros((0,self.distribution.dim))
        while samps.shape[0] != m:
            samps = self.bias_distribution.rvs(m, *args, **kwargs)
            # ratio = 
        return (samps, np.ones(m)/float(m))


#
# This file is part of TransportMaps.
#
# TransportMaps is free software: you can redistribute it and/or modify
# it under the terms of the LGNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# TransportMaps is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# LGNU Lesser General Public License for more details.
#
# You should have received a copy of the LGNU Lesser General Public License
# along with TransportMaps.  If not, see <http://www.gnu.org/licenses/>.
#
# Transport Maps Library
# Copyright (C) 2015-2016 Massachusetts Institute of Technology
# Uncertainty Quantification group
# Department of Aeronautics and Astronautics
#
# Author: Transport Map Team
# E-mail: tmteam@mit.edu
# Website: transport-maps.mit.edu
# Support: transport-maps.mit.edu/qa/
#

import numpy as np
import scipy.stats as stats

from TransportMaps.Defaults import Default_IsotropicIntegratedSquaredTriangularTransportMap, Default_IsotropicIntegratedExponentialTriangularTransportMap
from TransportMaps.Distributions.FrozenDistributions import StandardNormalDistribution
from TransportMaps.Distributions.TransportMapDistributions import PushForwardTransportMapDistribution, PullBackTransportMapDistribution
import pdb
# Tolerance for l1 optimization
tol_l1 = 1e-6

# Threshold for coefficient values after opt
l1_threshold = 1e-6

# params for l1 optimization
max_iter = 10
tol_soln = 1e-6
eps = 1e-3
lambda_list = np.logspace(-6,0,20)

def reweighted_l1(eta, pi, qtype, qparams, REG, tol, ders, max_iter, tol_soln, eps, dim, order, active_vars):

    # initialize flags for reweighted l1
    err_soln = 1
    iter_count = 0

    # Build the transport map (isotropic for each entry)
    # *********************************TESTING*****************************
    tm_approx = Default_IsotropicIntegratedSquaredTriangularTransportMap(
          dim, order, active_vars=active_vars)
    #tm_approx = Default_IsotropicIntegratedExponentialTriangularTransportMap(
    #      dim, order, active_vars=active_vars)
    # ******************************END TESTING*****************************

    # Construct density T_\sharp \pi
    tm_density = PushForwardTransportMapDistribution(tm_approx, pi)

    # set initial l1_weights
    REG['l1_weights'] = np.ones(tm_approx.n_coeffs)

    # Compute initial estimate for map and get weights
    solve = tm_density.minimize_kl_divergence(eta, qtype=qtype, qparams=qparams, \
            regularization=REG, tol=tol, ders=ders)
    coeffs = tm_density.coeffs

    while(err_soln > tol_soln and iter_count < max_iter):

        # Update weights based on coeffs
        weights = 1./(np.abs(coeffs) + eps)
        REG['l1_weights'] = weights

        # Compute new map and update weights
        solve = tm_density.minimize_kl_divergence(eta, qtype=qtype, \
            qparams=qparams, regularization=REG, tol=tol, ders=ders)
        new_coeffs = tm_density.coeffs

        # Compute err_soln
        err_soln = np.linalg.norm(new_coeffs - coeffs)
        coeffs = new_coeffs

        # Update counter
        iter_count = iter_count + 1

    return tm_approx

def coeff_to_activevars(approx_list, threshold, prev_active_vars):

    dim = len(approx_list)
    # add diagonal map variables to active variables
    active_vars = [[i] for i in range(dim)]

    # loop over dimension and extract active vars for each
    for i in range(dim):

        # Extract and threshold coeffs
        c_coeffs = approx_list[i].c.coeffs
        h_coeffs = approx_list[i].h.coeffs
        c_coeffs[np.abs(c_coeffs) < threshold] = 0
        h_coeffs[np.abs(h_coeffs) < threshold] = 0

        # Extract c and h multi_idxs
        c_idx = approx_list[i].c.get_multi_idxs()
        h_idx = approx_list[i].h.get_multi_idxs()

        for j_idx,j in enumerate(prev_active_vars[i]):

            # Check if each coefficient in c is nonzero and variable involved in term
            for k in range(len(c_idx)):
                try:
                    if c_idx[k][j_idx] != 0 and c_coeffs[k] != 0:
                        active_vars[i].append(j)
                        break
                except:
                    pdb.set_trace()

            # Check if each coefficient in c is nonzero and variable involved in term
            for k in range(len(h_idx)):
                if h_idx[k][j_idx] != 0 and h_coeffs[k] != 0:
                    active_vars[i].append(j)
                    break

        # sort and take unique elements in list of active_vars
        active_vars[i] = sorted(set(active_vars[i]))

    return active_vars

def regularized_map(eta, pi, data, qtype, qparams, dim, order, active_vars, tol, ders, REG):
    nsamps = data.shape[0]

    if REG == None:

        # Build the transport map (isotropic for each entry)
        tm_approx = Default_IsotropicIntegratedSquaredTriangularTransportMap(
                  dim, order, active_vars=active_vars)

        # Construct density T_\sharp \pi
        tm_density = PushForwardTransportMapDistribution(tm_approx, pi)

        # Set regularization in REG
        REG = {'type': None }
        solve = tm_density.minimize_kl_divergence(eta, qtype=qtype,
                                                  qparams=qparams,
                                                  regularization=REG,
                                                  tol=tol, ders=ders)

        coeff_iter = tm_density.coeffs

    elif REG['type'] == 'L1':
        # Set l1 tolerance
        REG['tol_l1'] = tol_l1
        # Set ders to be maximum of 1 with L1
        ders = np.min((ders, 1))
        # Create list to store approx_list and BIC values
        coeffs_list = []
        bic = []
        for lambda_t in lambda_list:
            # set regularization in REG
            REG['alpha'] = lambda_t
            if REG['reweight'] == None or REG['reweight'] == False:
                # Build the transport map (isotropic for each entry)
                tm_approx = Default_IsotropicIntegratedSquaredTriangularTransportMap(
                      dim, order, active_vars=active_vars)

                # Construct density T_\sharp \pi
                tm_density = PushForwardTransportMapDistribution(tm_approx, pi)

                # set regularization in REG
                REG['l1_weights'] = np.ones(tm_approx.n_coeffs)
                tm_density.minimize_kl_divergence(eta, qtype=qtype,
                                                          qparams=qparams,
                                                          regularization=REG,
                                                          tol=tol, ders=ders)
            elif REG['reweight'] == True:
                tm_approx = reweighted_l1(eta, pi, qtype, qparams, REG, tol, ders, max_iter, tol_soln, eps, dim, order, active_vars)

            coeffs_list.append(tm_approx.approx_list)
            coeff_iter = tm_approx.coeffs
            coeff_iter[abs(coeff_iter) < l1_threshold] = 0.0
            # Evaluate log-likelihood for data used to create map
            pull_S = PullBackTransportMapDistribution(tm_approx, eta)
            log_lik_trial = np.sum(pull_S.log_pdf(data))
            bic.append(np.log(nsamps)*np.count_nonzero(coeff_iter) - 2*log_lik_trial)

        idx = np.argmin(bic)
        coeffs_opt = coeffs_list[idx]
        new_active_vars = coeff_to_activevars(coeffs_opt, l1_threshold, tm_approx.active_vars)
        # Rerun transport map optimization (isotropic for each entry)
        print(new_active_vars)
        tm_approx = Default_IsotropicIntegratedSquaredTriangularTransportMap(
              dim, order, active_vars=new_active_vars)
        # Construct density T_\sharp \pi
        tm_density = PushForwardTransportMapDistribution(tm_approx, pi)
        REG = {'type': None }
        tm_density.minimize_kl_divergence(eta, qtype=qtype,
                                          qparams=qparams,
                                          regularization=REG,
                                          tol=tol, ders=ders)
    
    return tm_approx

if __name__ == "__main__":
    import GaussianGraphs as gg
    import ExtraDensities 
    dim = 12
    nsamps = 1000
    tol = 1e-6
    graph_obj = gg.ChainGraph(dim)
    data = graph_obj.rvs(nsamps)
    eta = StandardNormalDistribution(dim)
    # Gradient information
    ders = 1
    # Approximation orders
    order = 2
    qtype = 0
    qparams = nsamps
    active_vars=None
    # define base_density from samples
    pi = ExtraDensities.SamplesDensity(data)

    REG = {'type': 'L1', 'reweight': True}
    tm_approx = regularized_map(eta, pi, data, qtype, qparams, dim, order, active_vars, tol, ders, REG)
    print(tm_approx.coeffs)



#
# This file is part of TransportMaps.
#
# TransportMaps is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# TransportMaps is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with TransportMaps.  If not, see <http://www.gnu.org/licenses/>.
#
# Transport Maps Library
# Copyright (C) 2015-2018 Massachusetts Institute of Technology
# Uncertainty Quantification group
# Department of Aeronautics and Astronautics
#
# Author: Transport Map Team
# Website: transportmaps.mit.edu
# Support: transportmaps.mit.edu/qa/
#

import sys, getopt
import codecs
import numpy as np
import matplotlib.pyplot as plt
import matplotlib.dates as mdates
import matplotlib.cbook as cbook

def bytespdate2num(fmt, encoding='us-ascii'):
    strconverter = mdates.strpdate2num(fmt)
    def bytesconverter(b):
        s = b.decode(encoding)
        return strconverter(s)
    return bytesconverter

def usage():
    print('PlotExchanges.py --input-data=<fname>')

argv = sys.argv[1:]
IN_FNAME = None
try:
    opts, args = getopt.getopt(argv, "h", ["input-data="])
except getopt.GetoptError:
    usage()
    sys.exit(2)
for opt, arg in opts:
    if opt == '-h':
        usage()
        sys.exit()
    elif opt == '--input-data':
        IN_FNAME = arg
if None in [IN_FNAME]:
    usage()
    sys.exit(3)

with open(IN_FNAME,'r') as datafile:
    dates, exchanges, variations = \
        np.loadtxt(datafile, skiprows=2, usecols=[0,1,2], delimiter=',',
                   unpack=True, converters={0: bytespdate2num('%m/%d/%y')})

fig = plt.figure()
ax = fig.add_subplot(111)
ax.plot(dates, exchanges)
ax.xaxis_date()
ax.xaxis.set_major_formatter(mdates.DateFormatter('%m/%d/%Y'))
ax.set_xlabel("Day")
ax.set_ylabel("Exchange rate")
ax1 = ax.twiny()
ax1.set_xlabel("Number")
ax1.plot(exchanges,alpha=0)
fig.autofmt_xdate()

fig = plt.figure()
ax = fig.add_subplot(111)
ax.plot(dates,variations)
ax.xaxis_date()
ax.xaxis.set_major_formatter(mdates.DateFormatter('%m/%d/%Y'))
ax.set_xlabel("Day")
ax.set_ylabel("Variation")
ax1 = ax.twiny()
ax1.set_xlabel("Number")
ax1.plot(variations,alpha=0)
fig.autofmt_xdate()

plt.show(False)


import xun
import numpy as np
import matplotlib.pyplot as plt

# Add constant to vector
a = np.zeros(3)
b = xun.sum_const_vec(2.,a)
print(a)
print(b)

# Add constant to matrix
a2 = np.zeros((2,3))
b2 = xun.sum_const_mat(2.,a2)
print(a2)
print(b2)

# Sample Weibull (2,4)
a = 2.
b = 4.
def pdf(x):
    return a/b * (x/b)**(a-1.) * np.exp(-(x/b)**a)
n = 10000
c = xun.weird_samples(n)
xx = np.linspace(np.min(c),np.max(c),100)
plt.figure()
plt.hist(c,bins=20,normed=True,histtype='step')
plt.plot(xx,pdf(xx))
plt.show(False)


#
# This file is part of TransportMaps.
#
# TransportMaps is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# TransportMaps is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with TransportMaps.  If not, see <http://www.gnu.org/licenses/>.
#
# Transport Maps Library
# Copyright (C) 2015-2018 Massachusetts Institute of Technology
# Uncertainty Quantification group
# Department of Aeronautics and Astronautics
#
# Author: Transport Map Team
# Website: transportmaps.mit.edu
# Support: transportmaps.mit.edu/qa/
#

import logging
import unittest
import numpy as np
import numpy.random as npr
import numpy.linalg as npla

try:
    import mpi_map
    MPI_SUPPORT = True
except:
    MPI_SUPPORT = False

class TransportMap_DerivativeChecks(unittest.TestCase):

    def setUp(self):
        npr.seed(1)
        import TransportMaps.Distributions as DIST
        self.dim = 2
        self.order = 3
        self.fd_eps = 1e-6
        self.qtype = 3
        self.qpar = [5]*self.dim
        self.distribution = DIST.StandardNormalDistribution(self.dim)
        (x,w) = self.distribution.quadrature(self.qtype, self.qpar)
        self.x = x
        self.w = w
        self.y = np.zeros((1,self.dim))
        self.build_tm_approx()

    def test_grad_a(self):
        import TransportMaps.FiniteDifference as FD

        tm_approx = self.tm_approx
        coeffs = self.coeffs
        params = self.params
        x = self.x

        # Define transport map, gradient
        def tm(a, params_t):
            tm_approx.coeffs = a
            out = tm_approx.evaluate(x, params_t)
            return out
        def grad_a_tm(a, params_t):
            tm_approx.coeffs = a
            ncoeffs = tm_approx.n_coeffs
            grad_list = tm_approx.grad_a(x, params_t)
            out = np.zeros((len(x), tm_approx.dim, ncoeffs))
            start = 0
            for k, grad in enumerate(grad_list):
                stop = start + grad.shape[1]
                out[:,k,start:stop] = grad
                start = stop
            return out

        # Check gradient transport map
        flag = FD.check_grad_a(tm, grad_a_tm, coeffs, self.fd_eps,
                               params, verbose=False)
        self.assertTrue( flag )

    def test_hess_a(self):
        import TransportMaps.FiniteDifference as FD

        tm_approx = self.tm_approx
        coeffs = self.coeffs
        params = self.params
        x = self.x

        # Define transport gradient and hessian
        def grad_a_tm(a, params_t):
            tm_approx.coeffs = a
            ncoeffs = tm_approx.n_coeffs
            grad_list = tm_approx.grad_a(x, params_t)
            out = np.zeros((len(x), tm_approx.dim, ncoeffs))
            start = 0
            for k, grad in enumerate(grad_list):
                stop = start + grad.shape[1]
                out[:,k,start:stop] = grad
                start = stop
            return out
        def hess_a_tm(a, params_t):
            tm_approx.coeffs = a
            ncoeffs = tm_approx.n_coeffs
            hess_list = tm_approx.hess_a(x, params_t)
            out = np.zeros((len(x), tm_approx.dim, ncoeffs, ncoeffs))
            start = 0
            for k, hess in enumerate(hess_list):
                stop = start + hess.shape[1]
                out[:,k,start:stop,start:stop] = hess
                start = stop
            return out

        # Check Hessian transport map
        flag = FD.check_hess_a_from_grad_a(grad_a_tm, hess_a_tm, coeffs, self.fd_eps,
                                           params, verbose=False)
        self.assertTrue( flag )

    def test_action_hess_a(self):
        import TransportMaps.FiniteDifference as FD

        tm_approx = self.tm_approx
        coeffs = self.coeffs
        da = 1e-1 * npr.randn(coeffs.size)
        params = self.params
        x = self.x

        # Define transport gradient and hessian
        def hess_a_dot_da(a, da, params_t):
            tm_approx.coeffs = a
            ncoeffs = tm_approx.n_coeffs
            hess_list = tm_approx.hess_a(x, params_t)
            A = np.zeros((len(x), tm_approx.dim, ncoeffs, ncoeffs))
            start = 0
            for k, hess in enumerate(hess_list):
                stop = start + hess.shape[1]
                A[:,k,start:stop,start:stop] = hess
                start = stop
            return np.dot(A, da)
        def action_hess_a(a, da, params_t):
            tm_approx.coeffs = a
            ncoeffs = tm_approx.n_coeffs
            action_hess_a_list = tm_approx.action_hess_a(x, da, params_t)
            out = np.zeros((len(x), tm_approx.dim, ncoeffs))
            start = 0
            for k, aha in enumerate(action_hess_a_list):
                stop = start + aha.shape[1]
                out[:,k,start:stop] = aha
                start = stop
            return out

        ha_dot_da = hess_a_dot_da(coeffs, da, **params)
        aha = action_hess_a(coeffs, da, **params)
        self.assertTrue( np.allclose(ha_dot_da, aha) )

    def test_grad_a_partial_xd(self):
        import TransportMaps.FiniteDifference as FD

        tm_approx = self.tm_approx
        coeffs = self.coeffs
        params = self.params
        x = self.x

        # Define transport map, gradient
        def partial_xd_tm(a, params_t):
            tm_approx.coeffs = a
            out = tm_approx.partial_xd(x, params_t)
            return out
        def grad_a_partial_xd_tm(a, params_t):
            tm_approx.coeffs = a
            out = tm_approx.grad_a_partial_xd(x, params_t)
            return out

        # Check gradient transport map
        flag = FD.check_grad_a(partial_xd_tm, grad_a_partial_xd_tm,
                               coeffs, self.fd_eps,
                               params, verbose=False)
        self.assertTrue( flag )

    def test_grad_a_log_det_grad_x(self):
        import TransportMaps.FiniteDifference as FD

        tm_approx = self.tm_approx
        coeffs = self.coeffs
        params = self.params
        x = self.x

        # Define transport map, gradient
        def log_det_grad_x_tm(a, params_t):
            tm_approx.coeffs = a
            out = tm_approx.log_det_grad_x(x, params_t)
            return out
        def grad_a_log_det_grad_x_tm(a, params_t):
            tm_approx.coeffs = a
            out = tm_approx.grad_a_log_det_grad_x(x, params_t)
            return out

        # Check gradient transport map
        flag = FD.check_grad_a(log_det_grad_x_tm, grad_a_log_det_grad_x_tm,
                               coeffs, self.fd_eps,
                               params, verbose=False)
        self.assertTrue( flag )

    def test_hess_a_log_det_grad_x(self):
        import TransportMaps.FiniteDifference as FD

        tm_approx = self.tm_approx
        coeffs = self.coeffs
        params = self.params
        x = self.x

        # Define transport map, gradient
        def grad_a_log_det_grad_x_tm(a, params_t):
            tm_approx.coeffs = a
            out = tm_approx.grad_a_log_det_grad_x(x, params_t)
            return out
        def hess_a_log_det_grad_x_tm(a, params_t):
            tm_approx.coeffs = a
            out = tm_approx.hess_a_log_det_grad_x(x, params_t)
            return out

        # Check Hessian transport map
        flag = FD.check_hess_a_from_grad_a(grad_a_log_det_grad_x_tm,
                                           hess_a_log_det_grad_x_tm,
                                           coeffs, self.fd_eps,
                                           params, verbose=False)
        self.assertTrue( flag )

    def test_action_hess_a_log_det_grad_x(self):
        import TransportMaps.FiniteDifference as FD

        tm_approx = self.tm_approx
        coeffs = self.coeffs
        da = 1e-1 * npr.randn(coeffs.size)
        params = self.params
        x = self.x

        # Define transport gradient and hessian
        def hess_a_dot_da(a, da, params_t):
            tm_approx.coeffs = a
            haldgx = tm_approx.hess_a_log_det_grad_x(x, params_t)
            return np.dot(haldgx, da)
        def action_hess_a(a, da, params_t):
            tm_approx.coeffs = a
            return tm_approx.action_hess_a_log_det_grad_x(x, da, params_t)

        ha_dot_da = hess_a_dot_da(coeffs, da, **params)
        aha = action_hess_a(coeffs, da, **params)
        self.assertTrue( np.allclose(ha_dot_da, aha) )
        
    def test_inverse(self):
        tm_approx = self.tm_approx
        if hasattr(self,'coeffs'):
            coeffs = self.coeffs
            tm_approx.coeffs = coeffs
        y = self.y
        x = tm_approx.inverse(y)
        yrec = tm_approx.evaluate(x)
        self.assertTrue( np.allclose(y, yrec, rtol=1e-10, atol=1e-10) )

    def test_grad_a_inverse(self):
        import TransportMaps.FiniteDifference as FD

        tm_approx = self.tm_approx
        coeffs = self.coeffs
        params = self.params
        y = self.y

        # Define transport map, gradient
        def tm_inverse(a, params_t):
            tm_approx.coeffs = a
            out = tm_approx.inverse(y)
            return out
        def grad_a_tm_inverse(a, params_t):
            tm_approx.coeffs = a
            out = tm_approx.grad_a_inverse(y)
            return out

        # Check gradient transport map
        flag = FD.check_grad_a(tm_inverse, grad_a_tm_inverse, coeffs, self.fd_eps,
                               params, verbose=False)
        self.assertTrue( flag )

class OnTheFly_TM_DerivativeChecks( TransportMap_DerivativeChecks ):
    """ Special test definitions for evaluation on-the-fly
    """
    def test_grad_x(self):
        import TransportMaps.FiniteDifference as FD

        tm_approx = self.tm_approx
        if hasattr(self, 'coeffs'):
            tm_approx.coeffs = self.coeffs
        params = self.params
        x = self.x

        # Define transport map, gradient
        def tm(x, params_t=None):
            out = tm_approx.evaluate(x)
            return out
        def grad_x_tm(x, params_t=None):
            out = tm_approx.grad_x(x)
            return out

        # Check gradient transport map
        flag = FD.check_grad_x(tm, grad_x_tm, x, self.fd_eps,
                               params, verbose=False)
        self.assertTrue( flag )

    def test_grad_x_inverse(self):
        import TransportMaps.FiniteDifference as FD

        tm_approx = self.tm_approx
        if hasattr(self, 'coeffs'):
            tm_approx.coeffs = self.coeffs
        params = self.params
        y = self.y

        # Define transport map, gradient
        def tm(x, params_t=None):
            out = tm_approx.inverse(x)
            return out
        def grad_x_tm(x, params_t=None):
            out = tm_approx.grad_x_inverse(x)
            return out

        # Check gradient transport map
        flag = FD.check_grad_x(tm, grad_x_tm, y, self.fd_eps,
                               params, verbose=False)
        self.assertTrue( flag )

    def test_hess_x(self):
        import TransportMaps.FiniteDifference as FD

        tm_approx = self.tm_approx
        if hasattr(self, 'coeffs'):
            tm_approx.coeffs = self.coeffs
        params = self.params
        x = self.x

        # Define transport gradient and hessian
        def grad_x_tm(x, params_t=None):
            out = tm_approx.grad_x(x)
            return out
        def hess_x_tm(x, params_t=None):
            out = tm_approx.hess_x(x)
            return out

        # Check Hessian transport map
        flag = FD.check_hess_x_from_grad_x(grad_x_tm, hess_x_tm,
                                           x, self.fd_eps,
                                           params, verbose=False)
        self.assertTrue( flag )

    def test_action_hess_x(self):
        import TransportMaps.FiniteDifference as FD

        tm_approx = self.tm_approx
        if hasattr(self, 'coeffs'):
            tm_approx.coeffs = self.coeffs
        params = self.params
        x = self.x
        dx = 1e-1 * npr.randn(*(x.shape))

        # Define transport gradient and hessian
        A = tm_approx.hess_x(x)
        hx_dot_dx = np.einsum('...ijk,...k->...ij', A, dx)
        ahx = tm_approx.action_hess_x(x, dx)
        self.assertTrue( np.allclose(hx_dot_dx, ahx) )

    def test_grad_a_grad_x(self):
        import TransportMaps.FiniteDifference as FD

        tm_approx = self.tm_approx
        coeffs = self.coeffs
        params = self.params
        x = self.x

        # Define transport gradient and hessian
        def grad_a_tm(x, params_t):
            ncoeffs = tm_approx.n_coeffs
            grad_list = tm_approx.grad_a(x)
            out = np.zeros((len(x), ncoeffs, tm_approx.dim))
            start = 0
            for k, grad in enumerate(grad_list):
                stop = start + grad.shape[1]
                out[:,start:stop,k] = grad
                start = stop
            return out

        def grad_a_grad_x_tm(x, params_t):
            ncoeffs = tm_approx.n_coeffs
            grad_list = tm_approx.grad_a_grad_x(x)
            out = np.zeros((len(x), ncoeffs, tm_approx.dim, tm_approx.dim))
            start = 0
            for k, grad in enumerate(grad_list):
                stop = start + grad.shape[1]
                out[:,start:stop,k,:k+1] = grad
                start = stop
            return out

        # Check gradient transport map
        flag = FD.check_grad_x(grad_a_tm, grad_a_grad_x_tm, x, self.fd_eps,
                               params, verbose=False)
        self.assertTrue( flag )

    def test_grad_a_hess_x(self):
        import TransportMaps.FiniteDifference as FD

        tm_approx = self.tm_approx
        coeffs = self.coeffs
        params = self.params
        x = self.x

        # Define transport gradient and hessian
        def grad_a_grad_x_tm(x, params_t):
            ncoeffs = tm_approx.n_coeffs
            grad_list = tm_approx.grad_a_grad_x(x)
            out = np.zeros((len(x), ncoeffs, tm_approx.dim, tm_approx.dim))
            start = 0
            for k, grad in enumerate(grad_list):
                stop = start + grad.shape[1]
                out[:,start:stop,k,:k+1] = grad
                start = stop
            return out

        def grad_a_hess_x_tm(x, params_t):
            ncoeffs = tm_approx.n_coeffs
            hess_list = tm_approx.grad_a_hess_x(x)
            out = np.zeros((len(x), ncoeffs, tm_approx.dim, tm_approx.dim, tm_approx.dim))
            start = 0
            for k, grad in enumerate(hess_list):
                stop = start + grad.shape[1]
                out[:,start:stop,k,:k+1,:k+1] = grad
                start = stop
            return out

        # Check gradient transport map
        flag = FD.check_hess_x_from_grad_x(grad_a_grad_x_tm, grad_a_hess_x_tm, x, self.fd_eps,
                                           params, verbose=False)
        self.assertTrue( flag )

    def test_log_det_grad_x(self):
        tm_approx = self.tm_approx
        if hasattr(self, 'coeffs'):
            tm_approx.coeffs = self.coeffs
        params = self.params
        x = self.x

        # Compute \log\det\nabla_xT from \nabla_xT
        gx = tm_approx.grad_x(x)
        ldgx_from_gx = np.zeros(x.shape[0])
        for i in range(x.shape[0]):
            ldgx_from_gx[i] = np.log(npla.det(gx[i,:,:]))
        # Compute \log\det\nabla_xT directly
        ldgx = tm_approx.log_det_grad_x(x)

        # Check result
        flag = np.max(np.abs(ldgx - ldgx_from_gx)) < 1e-10
        self.assertTrue(flag)

    def test_grad_x_log_det_grad_x(self):
        import TransportMaps.FiniteDifference as FD

        tm_approx = self.tm_approx
        if hasattr(self, 'coeffs'):
            tm_approx.coeffs = self.coeffs
        params = self.params
        x = self.x

        # Define transport map, gradient
        def log_det_grad_x_tm(x, params_t=None):
            out = tm_approx.log_det_grad_x(x)
            return out
        def grad_x_log_det_grad_x_tm(x, params_t=None):
            out = tm_approx.grad_x_log_det_grad_x(x)
            return out

        # Check gradient transport map
        flag = FD.check_grad_x(log_det_grad_x_tm, grad_x_log_det_grad_x_tm,
                               x, self.fd_eps,
                               params, verbose=False)
        self.assertTrue( flag )

    def test_hess_x_log_det_grad_x(self):
        import TransportMaps.FiniteDifference as FD

        tm_approx = self.tm_approx
        if hasattr(self,'coeffs'):
            coeffs = self.coeffs
            tm_approx.coeffs = coeffs
        params = self.params
        x = self.x

        # Define transport map, gradient
        def grad_x_log_det_grad_x_tm(x, params_t=None):
            out = tm_approx.grad_x_log_det_grad_x(x)
            return out
        def hess_x_log_det_grad_x_tm(x, params_t=None):
            out = tm_approx.hess_x_log_det_grad_x(x)
            return out

        # Check gradient transport map
        flag = FD.check_hess_x_from_grad_x(grad_x_log_det_grad_x_tm,
                                           hess_x_log_det_grad_x_tm,
                                           x, self.fd_eps,
                                           params, verbose=False)
        self.assertTrue( flag )

    def test_action_hess_x_log_det_grad_x(self):
        import TransportMaps.FiniteDifference as FD

        tm_approx = self.tm_approx
        if hasattr(self,'coeffs'):
            coeffs = self.coeffs
            tm_approx.coeffs = coeffs
        params = self.params
        x = self.x
        dx = 1e-1 * npr.randn(*(x.shape))

        # Define transport map, gradient
        A = tm_approx.hess_x_log_det_grad_x(x)
        hx_dot_dx = np.einsum('...jk,...k->...j', A, dx)
        ahx = tm_approx.action_hess_x_log_det_grad_x(x, dx)
        self.assertTrue( np.allclose(hx_dot_dx, ahx) )

    def test_log_det_grad_x_inverse(self):
        tm_approx = self.tm_approx
        if hasattr(self, 'coeffs'):
            tm_approx.coeffs = self.coeffs
        params = self.params
        y = self.y

        # Compute \log\det\nabla_xT from \nabla_xT
        gx = tm_approx.grad_x_inverse(y)
        ldgx_from_gx = np.zeros(y.shape[0])
        for i in range(y.shape[0]):
            ldgx_from_gx[i] = np.log(npla.det(gx[i,:,:]))
        # Compute \log\det\nabla_xT directly
        ldgx = tm_approx.log_det_grad_x_inverse(y)

        # Check result
        flag = np.max(np.abs(ldgx - ldgx_from_gx)) < 1e-10
        self.assertTrue(flag)

    # def test_grad_x_log_det_grad_x_inverse(self):
    #     import TransportMaps.FiniteDifference as FD

    #     tm_approx = self.tm_approx
    #     tm_approx.coeffs = self.coeffs
    #     params = self.params
    #     x = self.x

    #     # Define transport map, gradient
    #     def log_det_grad_x_tm(x, params_t=None):
    #         out = tm_approx.log_det_grad_x_inverse(x)
    #         return out
    #     def grad_x_log_det_grad_x_tm(x, params_t=None):
    #         out = tm_approx.grad_x_log_det_grad_x_inverse(x)
    #         return out

    #     # Check gradient transport map
    #     flag = FD.check_grad_x(log_det_grad_x_tm, grad_x_log_det_grad_x_tm,
    #                            x, self.fd_eps,
    #                            params, verbose=False)
    #     self.assertTrue( flag )

    # def test_hess_x_log_det_grad_x_inverse(self):
    #     import TransportMaps.FiniteDifference as FD

    #     tm_approx = self.tm_approx
    #     coeffs = self.coeffs
    #     tm_approx.coeffs = coeffs
    #     params = self.params
    #     x = self.x

    #     # Define transport map, gradient
    #     def grad_x_log_det_grad_x_tm(x, params_t=None):
    #         out = tm_approx.grad_x_log_det_grad_x_inverse(x)
    #         return out
    #     def hess_x_log_det_grad_x_tm(x, params_t=None):
    #         out = tm_approx.hess_x_log_det_grad_x_inverse(x)
    #         return out

    #     # Check gradient transport map
    #     flag = FD.check_hess_x_from_grad_x(grad_x_log_det_grad_x_tm,
    #                                        hess_x_log_det_grad_x_tm,
    #                                        x, self.fd_eps,
    #                                        params, verbose=False)
    #     self.assertTrue( flag )

    def test_grad_a_hess_x_log_det_grad_x(self):
        import TransportMaps.FiniteDifference as FD

        tm_approx = self.tm_approx
        coeffs = self.coeffs
        tm_approx.coeffs = coeffs
        params = self.params
        x = self.x

        # Define transport map, gradient
        def hess_x_log_det_grad_x_tm(a, params_t):
            tm_approx.coeffs = a
            out = tm_approx.hess_x_log_det_grad_x(x)
            return out
        def grad_a_hess_x_log_det_grad_x_tm(a, params_t):
            tm_approx.coeffs = a
            out = tm_approx.grad_a_hess_x_log_det_grad_x(x)
            return out

        # Check gradient transport map
        flag = FD.check_grad_a(hess_x_log_det_grad_x_tm, grad_a_hess_x_log_det_grad_x_tm,
                               coeffs, self.fd_eps, params, end=False, verbose=False)
        self.assertTrue( flag )

class OnTheFly_InverseTM_DerivativeChecks( TransportMap_DerivativeChecks ):
    def test_grad_x_inverse(self):
        import TransportMaps.FiniteDifference as FD

        tm_approx = self.tm_approx
        if hasattr(self, 'coeffs'):
            tm_approx.coeffs = self.coeffs
        params = self.params
        x = self.x

        # Define transport map, gradient
        def tm(x, params_t=None):
            out = tm_approx.inverse(x)
            return out
        def grad_x_tm(x, params_t=None):
            out = tm_approx.grad_x_inverse(x)
            return out

        # Check gradient transport map
        flag = FD.check_grad_x(tm, grad_x_tm, x, self.fd_eps,
                               params, verbose=False)
        self.assertTrue( flag )

    def test_grad_a_inverse(self):
        import TransportMaps.FiniteDifference as FD

        tm_approx = self.tm_approx
        coeffs = self.coeffs
        params = self.params
        x = self.x

        # Define transport map, gradient
        def tm(a, params_t):
            tm_approx.coeffs = a
            out = tm_approx.inverse(x)
            return out
        def grad_a_tm(a, params_t):
            tm_approx.coeffs = a
            ncoeffs = tm_approx.n_coeffs
            grad_list = tm_approx.grad_a_inverse(x)
            out = np.zeros((len(x), tm_approx.dim, ncoeffs))
            start = 0
            for k, grad in enumerate(grad_list):
                stop = start + grad.shape[1]
                out[:,k,start:stop] = grad
                start = stop
            return out

        # Check gradient transport map
        flag = FD.check_grad_a(tm, grad_a_tm, coeffs, self.fd_eps,
                               params, verbose=False)
        self.assertTrue( flag )

    def test_grad_a(self):
        import TransportMaps.FiniteDifference as FD

        tm_approx = self.tm_approx
        coeffs = self.coeffs
        params = self.params
        x = self.x

        # Define transport map, gradient
        def tm_inverse(a, params_t):
            tm_approx.coeffs = a
            out = tm_approx.evaluate(x)
            return out
        def grad_a_tm_inverse(a, params_t):
            tm_approx.coeffs = a
            out = tm_approx.grad_a(x)
            return out

        # Check gradient transport map
        flag = FD.check_grad_a(tm_inverse, grad_a_tm_inverse, coeffs, self.fd_eps,
                               params, verbose=False)
        self.assertTrue( flag )

    @unittest.skip("Not Implemented")
    def test_hess_a(self):
        pass

    @unittest.skip("Not Implemented")
    def test_action_hess_a(self):
        pass

    @unittest.skip("Not Implemented")
    def test_grad_a_log_det_grad_x(self):
        pass

    @unittest.skip("Not Implemented")
    def test_hess_a_log_det_grad_x(self):
        pass

    @unittest.skip("Not Implemented")
    def test_action_hess_a_log_det_grad_x(self):
        pass

    @unittest.skip("Not Implemented")
    def test_grad_a_partial_xd(self):
        pass

    @unittest.skip("Not Implemented")
    def test_grad_a_hess_x_log_det_grad_x(self):
        pass

class OnTheFly_CompositeTM_DerivativeChecks( OnTheFly_TM_DerivativeChecks ):
    @unittest.skip("Not Implemented")
    def test_grad_a(self):
        pass

    @unittest.skip("Not Implemented")
    def test_hess_a(self):
        pass

    @unittest.skip("Not Implemented")
    def test_action_hess_a(self):
        pass

    @unittest.skip("Not Implemented")
    def test_grad_a_inverse(self):
        pass

    @unittest.skip("Not Implemented")
    def test_grad_a_grad_x(self):
        pass

    @unittest.skip("Not Implemented")
    def test_grad_a_hess_x(self):
        pass

    @unittest.skip("Not Implemented")
    def test_grad_a_log_det_grad_x(self):
        pass

    @unittest.skip("Not Implemented")
    def test_hess_a_log_det_grad_x(self):
        pass

    @unittest.skip("Not Implemented")
    def test_action_hess_a_log_det_grad_x(self):
        pass

    @unittest.skip("Not Implemented")
    def test_grad_a_partial_xd(self):
        pass

    @unittest.skip("Not Implemented")
    def test_grad_a_hess_x_log_det_grad_x(self):
        pass

class Precomp_TM_DerivativeChecks( TransportMap_DerivativeChecks ):
    """ Special test definitions with precomputed Vandermonde matrices.
    """
    def test_grad_x(self):
        import TransportMaps.FiniteDifference as FD

        tm_approx = self.tm_approx
        if hasattr(self, 'coeffs'):
            tm_approx.coeffs = self.coeffs
        params = self.params
        x = self.x

        # Define transport map, gradient
        def tm(x, params_t=None):
            out = tm_approx.evaluate(x)
            return out
        def grad_x_tm(x, params_t=None):
            out = tm_approx.grad_x(x, params_t)
            return out

        # Check gradient transport map
        flag = FD.check_grad_x(tm, grad_x_tm, x, self.fd_eps,
                               params, verbose=False)
        self.assertTrue( flag )

    def test_hess_x(self):
        import TransportMaps.FiniteDifference as FD

        tm_approx = self.tm_approx
        if hasattr(self, 'coeffs'):
            tm_approx.coeffs = self.coeffs
        params = self.params
        x = self.x

        # Define transport gradient and hessian
        def grad_x_tm(x, params_t=None):
            out = tm_approx.grad_x(x)
            return out
        def hess_x_tm(x, params_t=None):
            out = tm_approx.hess_x(x, params_t)
            return out

        # Check Hessian transport map
        flag = FD.check_hess_x_from_grad_x(grad_x_tm, hess_x_tm,
                                           x, self.fd_eps,
                                           params, verbose=False)
        self.assertTrue( flag )

    def test_action_hess_x(self):
        import TransportMaps.FiniteDifference as FD

        tm_approx = self.tm_approx
        if hasattr(self, 'coeffs'):
            tm_approx.coeffs = self.coeffs
        params = self.params
        x = self.x
        dx = 1e-1 * npr.randn(*(x.shape))

        # Define transport gradient and hessian
        A = tm_approx.hess_x(x)
        hx_dot_dx = np.einsum('...ijk,...k->...ij', A, dx)
        ahx = tm_approx.action_hess_x(x, dx)
        self.assertTrue( np.allclose(hx_dot_dx, ahx) )

    def test_grad_x_log_det_grad_x(self):
        import TransportMaps.FiniteDifference as FD

        tm_approx = self.tm_approx
        if hasattr(self, 'coeffs'):
            tm_approx.coeffs = self.coeffs
        params = self.params
        x = self.x

        # Define transport map, gradient
        def log_det_grad_x_tm(x, params_t=None):
            out = tm_approx.log_det_grad_x(x)
            return out
        def grad_x_log_det_grad_x_tm(x, params_t=None):
            out = tm_approx.grad_x_log_det_grad_x(x, params_t)
            return out

        # Check gradient transport map
        flag = FD.check_grad_x(log_det_grad_x_tm, grad_x_log_det_grad_x_tm,
                               x, self.fd_eps,
                               params, verbose=False)
        self.assertTrue( flag )

    def test_hess_x_log_det_grad_x(self):
        import TransportMaps.FiniteDifference as FD

        tm_approx = self.tm_approx
        coeffs = self.coeffs
        tm_approx.coeffs = coeffs
        params = self.params
        x = self.x

        # Define transport map, gradient
        def grad_x_log_det_grad_x_tm(x, params_t=None):
            out = tm_approx.grad_x_log_det_grad_x(x)
            return out
        def hess_x_log_det_grad_x_tm(x, params_t=None):
            out = tm_approx.hess_x_log_det_grad_x(x, params_t)
            return out

        # Check gradient transport map
        flag = FD.check_hess_x_from_grad_x(grad_x_log_det_grad_x_tm,
                                           hess_x_log_det_grad_x_tm,
                                           x, self.fd_eps,
                                           params, verbose=False)
        self.assertTrue( flag )

    def test_action_hess_x_log_det_grad_x(self):
        import TransportMaps.FiniteDifference as FD

        tm_approx = self.tm_approx
        coeffs = self.coeffs
        tm_approx.coeffs = coeffs
        params = self.params
        x = self.x
        dx = 1e-1 * npr.randn(*(x.shape))

        # Define transport map, gradient
        A = tm_approx.hess_x_log_det_grad_x(x)
        hx_dot_dx = np.einsum('...jk,...k->...j', A, dx)
        ahx = tm_approx.action_hess_x_log_det_grad_x(x, dx)
        self.assertTrue( np.allclose(hx_dot_dx, ahx) )

class PrecompUni_TM_DerivativeChecks( Precomp_TM_DerivativeChecks ):
    """ Precompute uni-variate Vandermonde matrices
    """
    def setUp(self):
        super(PrecompUni_TM_DerivativeChecks,self).setUp()
        x = self.x
        self.params['params_t'] = {'components': [{} for i in range(self.dim)]}
        self.tm_approx.precomp_evaluate(x, self.params['params_t'])
        self.tm_approx.precomp_partial_xd(x, self.params['params_t'])
        self.tm_approx.precomp_grad_x(x, self.params['params_t'])
        self.tm_approx.precomp_hess_x(x, self.params['params_t'])
        self.tm_approx.precomp_grad_x_partial_xd(x, self.params['params_t'])
        self.tm_approx.precomp_hess_x_partial_xd(x, self.params['params_t'])

class PrecompMulti_TM_DerivativeChecks( Precomp_TM_DerivativeChecks ):
    """ Precompute multi-variate Vandermonde matrices
    """
    def setUp(self):
        super(PrecompMulti_TM_DerivativeChecks,self).setUp()
        x = self.x
        self.params['params_t'] = {'components': [{} for i in range(self.dim)]}
        self.tm_approx.precomp_evaluate(x, self.params['params_t'],
                                        precomp_type='multi')
        self.tm_approx.precomp_partial_xd(x, self.params['params_t'],
                                        precomp_type='multi')
        self.tm_approx.precomp_grad_x(x, self.params['params_t'],
                                      precomp_type='multi')
        self.tm_approx.precomp_hess_x(x, self.params['params_t'],
                                      precomp_type='multi')
        self.tm_approx.precomp_grad_x_partial_xd(x, self.params['params_t'],
                                                 precomp_type='multi')
        self.tm_approx.precomp_hess_x_partial_xd(x, self.params['params_t'],
                                                 precomp_type='multi')

class LinearTM( object ):
    def build_tm_approx(self):
        import TransportMaps.Maps as MAPS
        ct = npr.randn(self.dim)
        s = npr.randn(self.dim, self.dim)
        s = np.dot(s, s.T)
        lt = npla.cholesky(s)
        self.tm_approx = MAPS.LinearTransportMap(ct, lt)
        self.params = {}
        
class IntegratedExponentialTTM( object ):
    """ Integrated exponential approximation
    """
    def build_tm_approx(self):
        import SpectralToolbox.Spectral1D as S1D
        import TransportMaps as TM
        import TransportMaps.Functionals as FUNC
        import TransportMaps.Maps as MAPS
        # Build the transport map (isotropic for each entry)
        self.tm_approx = TM.Default_IsotropicIntegratedExponentialTriangularTransportMap(
            self.dim, self.order, span='full', btype='poly', common_basis_flag=False)
        self.params = {}
        self.params['params_t'] = None
        # self.coeffs = np.zeros(self.tm_approx.n_coeffs)
        self.coeffs = npr.randn(self.tm_approx.n_coeffs) / 10.

class CommonBasisIntegratedExponentialTTM( object ):
    """ Integrated exponential approximation
    """
    def build_tm_approx(self):
        import SpectralToolbox.Spectral1D as S1D
        import TransportMaps as TM
        import TransportMaps.Functionals as FUNC
        import TransportMaps.Maps as MAPS
        # Build the transport map (isotropic for each entry)
        self.tm_approx = TM.Default_IsotropicIntegratedExponentialTriangularTransportMap(
            self.dim, self.order, span='full', btype='poly', common_basis_flag=True)
        self.params = {}
        self.params['params_t'] = None
        # self.coeffs = np.zeros(self.tm_approx.n_coeffs)
        self.coeffs = npr.randn(self.tm_approx.n_coeffs) / 10.

class TotOrdIntegratedExponentialTTM( object ):
    """ Integrated exponential approximation
    """
    def build_tm_approx(self):
        import SpectralToolbox.Spectral1D as S1D
        import TransportMaps as TM
        import TransportMaps.Functionals as FUNC
        import TransportMaps.Maps as MAPS
        # Build the transport map (isotropic for each entry)
        self.tm_approx = TM.Default_IsotropicIntegratedExponentialTriangularTransportMap(
            self.dim, self.order, span='total', btype='poly', common_basis_flag=False)
        self.params = {}
        self.params['params_t'] = None
        # self.coeffs = np.zeros(self.tm_approx.n_coeffs)
        self.coeffs = npr.randn(self.tm_approx.n_coeffs) / 10.

class CommonBasisTotOrdIntegratedExponentialTTM( object ):
    """ Integrated exponential approximation
    """
    def build_tm_approx(self):
        import SpectralToolbox.Spectral1D as S1D
        import TransportMaps as TM
        import TransportMaps.Functionals as FUNC
        import TransportMaps.Maps as MAPS
        # Build the transport map (isotropic for each entry)
        self.tm_approx = TM.Default_IsotropicIntegratedExponentialTriangularTransportMap(
            self.dim, self.order, span='total', btype='poly', common_basis_flag=False) 
        self.params = {}
        self.params['params_t'] = None
        # self.coeffs = np.zeros(self.tm_approx.n_coeffs)
        self.coeffs = npr.randn(self.tm_approx.n_coeffs) / 10.

class IntegratedSquaredTTM( object ):
    """ Integrated squared approximation
    """
    def build_tm_approx(self):
        import SpectralToolbox.Spectral1D as S1D
        import TransportMaps as TM
        import TransportMaps.Functionals as FUNC
        import TransportMaps.Maps as MAPS
        # Build the transport map (isotropic for each entry)
        self.tm_approx = TM.Default_IsotropicIntegratedSquaredTriangularTransportMap(
            self.dim, self.order, span='full', btype='poly', common_basis_flag=False)
        self.params = {}
        self.params['params_t'] = None
        coeffs = self.tm_approx.get_identity_coeffs()
        coeffs += npr.randn(len(coeffs)) / 10.
        self.coeffs = np.hstack(coeffs)

class CommonBasisIntegratedSquaredTTM( object ):
    """ Integrated squared approximation
    """
    def build_tm_approx(self):
        import SpectralToolbox.Spectral1D as S1D
        import TransportMaps as TM
        import TransportMaps.Functionals as FUNC
        import TransportMaps.Maps as MAPS
        # Build the transport map (isotropic for each entry)
        self.tm_approx = TM.Default_IsotropicIntegratedSquaredTriangularTransportMap(
            self.dim, self.order, span='full', btype='poly', common_basis_flag=True)
        self.params = {}
        self.params['params_t'] = None
        coeffs = self.tm_approx.get_identity_coeffs()
        coeffs += npr.randn(len(coeffs)) / 10.
        self.coeffs = coeffs

class TotOrdIntegratedSquaredTTM( object ):
    """ Integrated squared approximation
    """
    def build_tm_approx(self):
        import SpectralToolbox.Spectral1D as S1D
        import TransportMaps as TM
        import TransportMaps.Functionals as FUNC
        import TransportMaps.Maps as MAPS
        # Build the transport map (isotropic for each entry)
        self.tm_approx = TM.Default_IsotropicIntegratedSquaredTriangularTransportMap(
            self.dim, self.order, span='total', btype='poly', common_basis_flag=False)
        self.params = {}
        self.params['params_t'] = None
        coeffs = self.tm_approx.get_identity_coeffs()
        coeffs += npr.randn(len(coeffs)) / 10.
        self.coeffs = coeffs

class CommonBasisTotOrdIntegratedSquaredTTM( object ):
    """ Integrated squared approximation
    """
    def build_tm_approx(self):
        import SpectralToolbox.Spectral1D as S1D
        import TransportMaps as TM
        import TransportMaps.Functionals as FUNC
        import TransportMaps.Maps as MAPS
        # Build the transport map (isotropic for each entry)
        self.tm_approx = TM.Default_IsotropicIntegratedSquaredTriangularTransportMap(
            self.dim, self.order, span='total', btype='poly', common_basis_flag=True)
        self.params = {}
        self.params['params_t'] = None
        coeffs = self.tm_approx.get_identity_coeffs()
        coeffs += npr.randn(len(coeffs)) / 10.
        self.coeffs = coeffs

# INTEGRATES SQUARED RBF

class IntegratedSquaredRBFTTM( object ):
    """ Integrated squared approximation
    """
    def build_tm_approx(self):
        import SpectralToolbox.Spectral1D as S1D
        import TransportMaps as TM
        import TransportMaps.Functionals as FUNC
        import TransportMaps.Maps as MAPS
        # Build the transport map (isotropic for each entry)
        self.tm_approx = TM.Default_IsotropicIntegratedSquaredTriangularTransportMap(
            self.dim, self.order, span='full', btype='rbf', common_basis_flag=False)
        self.params = {}
        self.params['params_t'] = None
        coeffs = []
        for approx in self.tm_approx.approx_list:
            coeffs.append( npr.random(approx.c.n_coeffs)/10. )
            coeffs.append( npr.random(approx.h.n_coeffs) )
        self.coeffs = np.hstack(coeffs)

class CommonBasisIntegratedSquaredRBFTTM( object ):
    """ Integrated squared approximation
    """
    def build_tm_approx(self):
        import SpectralToolbox.Spectral1D as S1D
        import TransportMaps as TM
        import TransportMaps.Functionals as FUNC
        import TransportMaps.Maps as MAPS
        # Build the transport map (isotropic for each entry)
        self.tm_approx = TM.Default_IsotropicIntegratedSquaredTriangularTransportMap(
            self.dim, self.order, span='full', btype='rbf', common_basis_flag=True)
        self.params = {}
        self.params['params_t'] = None
        self.coeffs = npr.randn(self.tm_approx.n_coeffs)

class TotOrdIntegratedSquaredRBFTTM( object ):
    """ Integrated exponential approximation
    """
    def build_tm_approx(self):
        import SpectralToolbox.Spectral1D as S1D
        import TransportMaps as TM
        import TransportMaps.Functionals as FUNC
        import TransportMaps.Maps as MAPS
        # Build the transport map (isotropic for each entry)
        self.tm_approx = TM.Default_IsotropicIntegratedSquaredTriangularTransportMap(
            self.dim, self.order, span='total', btype='rbf', common_basis_flag=False)
        self.params = {}
        self.params['params_t'] = None
        coeffs = []
        for approx in self.tm_approx.approx_list:
            coeffs.append( npr.random(approx.c.n_coeffs)/10. )
            coeffs.append( npr.random(approx.h.n_coeffs) )
        self.coeffs = np.hstack(coeffs)

class CommonBasisTotOrdIntegratedSquaredRBFTTM( object ):
    """ Integrated exponential approximation
    """
    def build_tm_approx(self):
        import SpectralToolbox.Spectral1D as S1D
        import TransportMaps as TM
        import TransportMaps.Functionals as FUNC
        import TransportMaps.Maps as MAPS
        # Build the transport map (isotropic for each entry)
        self.tm_approx = TM.Default_IsotropicIntegratedSquaredTriangularTransportMap(
            self.dim, self.order, span='total', btype='rbf', common_basis_flag=True)
        self.params = {}
        self.params['params_t'] = None
        coeffs = []
        for approx in self.tm_approx.approx_list:
            coeffs.append( npr.random(approx.c.n_coeffs)/10. )
            coeffs.append( npr.random(approx.h.n_coeffs) )
        self.coeffs = np.hstack(coeffs)

# LINEAR SPAN
        
class LinearSpanTTM( object ):
    """ Linear span approximation
    """
    def build_tm_approx(self):
        import SpectralToolbox.Spectral1D as S1D
        import TransportMaps as TM
        import TransportMaps.Functionals as FUNC
        import TransportMaps.Maps as MAPS
        self.tm_approx = TM.Default_IsotropicLinearSpanTriangularTransportMap(
            self.dim, self.order, span='full', btype='poly', common_basis_flag=False)
        self.params = {}
        self.params['params_t'] = None
        # Set coefficients for linear map
        self.coeffs = np.zeros(self.tm_approx.n_coeffs)
        all_midxs = []
        for tm_comp in self.tm_approx.approx_list:
            all_midxs.extend( tm_comp.multi_idxs )
        for d in range(self.tm_approx.dim):
            idx = all_midxs.index( tuple([0]*d + [1]) )
            self.coeffs[idx] = 1.

class CommonBasisLinearSpanTTM( object ):
    """ Linear span approximation
    """
    def build_tm_approx(self):
        import SpectralToolbox.Spectral1D as S1D
        import TransportMaps as TM
        import TransportMaps.Functionals as FUNC
        import TransportMaps.Maps as MAPS
        self.tm_approx = TM.Default_IsotropicLinearSpanTriangularTransportMap(
            self.dim, self.order, span='full', btype='poly', common_basis_flag=True)
        self.params = {}
        self.params['params_t'] = None
        # Set coefficients for linear map
        self.coeffs = np.zeros(self.tm_approx.n_coeffs)
        all_midxs = []
        for tm_comp in self.tm_approx.approx_list:
            all_midxs.extend( tm_comp.multi_idxs )
        for d in range(self.tm_approx.dim):
            idx = all_midxs.index( tuple([0]*d + [1]) )
            self.coeffs[idx] = 1.

class TotOrdLinearSpanTTM( object ):
    """ Linear span approximation
    """
    def build_tm_approx(self):
        import SpectralToolbox.Spectral1D as S1D
        import TransportMaps as TM
        import TransportMaps.Functionals as FUNC
        import TransportMaps.Maps as MAPS
        self.tm_approx = TM.Default_IsotropicLinearSpanTriangularTransportMap(
            self.dim, self.order, span='total', btype='poly', common_basis_flag=False)
        self.params = {}
        self.params['params_t'] = None
        # Set coefficients for linear map
        self.coeffs = np.zeros(self.tm_approx.n_coeffs)
        all_midxs = []
        for tm_comp in self.tm_approx.approx_list:
            all_midxs.extend( tm_comp.multi_idxs )
        for d in range(self.tm_approx.dim):
            idx = all_midxs.index( tuple([0]*d + [1]) )
            self.coeffs[idx] = 1.

class CommonBasisTotOrdLinearSpanTTM( object ):
    """ Linear span approximation
    """
    def build_tm_approx(self):
        import SpectralToolbox.Spectral1D as S1D
        import TransportMaps as TM
        import TransportMaps.Functionals as FUNC
        import TransportMaps.Maps as MAPS
        self.tm_approx = TM.Default_IsotropicLinearSpanTriangularTransportMap(
            self.dim, self.order, span='total', btype='poly', common_basis_flag=True)
        self.params = {}
        self.params['params_t'] = None
        # Set coefficients for linear map
        self.coeffs = np.zeros(self.tm_approx.n_coeffs)
        all_midxs = []
        for tm_comp in self.tm_approx.approx_list:
            all_midxs.extend( tm_comp.multi_idxs )
        for d in range(self.tm_approx.dim):
            idx = all_midxs.index( tuple([0]*d + [1]) )
            self.coeffs[idx] = 1.

#
# INVERSE MAP
#
class InverseCommonBasisTotOrdIntegratedExponentialTTM( object ):
    def build_tm_approx(self):
        import SpectralToolbox.Spectral1D as S1D
        import TransportMaps as TM
        import TransportMaps.Functionals as FUNC
        import TransportMaps.Maps as MAPS
        # Build the transport map (isotropic for each entry)
        approx_list = []
        active_vars = []
        c_basis_list = [S1D.HermiteProbabilistsPolynomial() for i in range(self.dim)]
        e_basis_list = [S1D.ConstantExtendedHermiteProbabilistsFunction(self.order)
                        for i in range(self.dim)]
        for i in range(self.dim):
            c_orders_list = ([self.order] * i) + [0]
            c_approx = FUNC.LinearSpanApproximation(c_basis_list[:i+1], spantype='total',
                                                             order_list=c_orders_list)
            e_orders_list = [self.order] * (i+1)
            e_approx = FUNC.LinearSpanApproximation(e_basis_list[:i+1], spantype='total',
                                                             order_list=e_orders_list)
            approx = FUNC.MonotonicIntegratedExponentialApproximation(c_approx, e_approx)
            approx_list.append( approx )
            active_vars.append( range(i+1) )
        tm_approx = MAPS.CommonBasisIntegratedExponentialTriangularTransportMap(active_vars,
                                                                                approx_list)
        self.tm_approx = MAPS.InverseTransportMap( tm_approx )
        self.params = {}
        self.params['params_t'] = None
        # self.coeffs = np.zeros(self.tm_approx.n_coeffs)
        self.coeffs = npr.randn(self.tm_approx.n_coeffs) / 10.

class InverseCommonBasisTotOrdIntegratedSquaredTTM( object ):
    def build_tm_approx(self):
        import SpectralToolbox.Spectral1D as S1D
        import TransportMaps as TM
        import TransportMaps.Functionals as FUNC
        import TransportMaps.Maps as MAPS
        # Build the transport map (isotropic for each entry)
        approx_list = []
        active_vars = []
        c_basis_list = [S1D.HermiteProbabilistsPolynomial() for i in range(self.dim)]
        e_basis_list = [S1D.ConstantExtendedHermiteProbabilistsFunction(self.order)
                        for i in range(self.dim)]
        for i in range(self.dim):
            c_orders_list = ([self.order] * i) + [0]
            c_approx = FUNC.LinearSpanApproximation(c_basis_list[:i+1], spantype='total',
                                                             order_list=c_orders_list)
            e_orders_list = [self.order] * (i+1)
            e_approx = FUNC.LinearSpanApproximation(e_basis_list[:i+1], spantype='total',
                                                             order_list=e_orders_list)
            approx = FUNC.MonotonicIntegratedSquaredApproximation(c_approx, e_approx)
            approx_list.append( approx )
            active_vars.append( range(i+1) )
        tm_approx = MAPS.CommonBasisIntegratedSquaredTriangularTransportMap(active_vars,
                                                                                approx_list)
        self.tm_approx = MAPS.InverseTransportMap( tm_approx )
        self.params = {}
        self.params['params_t'] = None
        # self.coeffs = np.zeros(self.tm_approx.n_coeffs)
        self.coeffs = npr.randn(self.tm_approx.n_coeffs) / 10.
        
#
# COMPOSITE MAP
#
class CompositeCommonBasisTotOrdIntegratedExponentialTTM( object ):
    def build_tm_approx(self):
        import SpectralToolbox.Spectral1D as S1D
        import TransportMaps as TM
        import TransportMaps.Functionals as FUNC
        import TransportMaps.Maps as MAPS
        tm_approx_list = []
        for i in range(2):
            # Build the transport map (isotropic for each entry)
            approx_list = []
            active_vars = []
            c_basis_list = [S1D.HermiteProbabilistsPolynomial() for i in range(self.dim)]
            e_basis_list = [S1D.ConstantExtendedHermiteProbabilistsFunction(self.order)
                            for i in range(self.dim)]
            for i in range(self.dim):
                c_orders_list = ([self.order] * i) + [0]
                c_approx = FUNC.LinearSpanApproximation(
                    c_basis_list[:i+1], spantype='total', order_list=c_orders_list)
                e_orders_list = [self.order] * (i+1)
                e_approx = FUNC.LinearSpanApproximation(
                    e_basis_list[:i+1], spantype='total', order_list=e_orders_list)
                approx = FUNC.MonotonicIntegratedExponentialApproximation(
                    c_approx, e_approx)
                approx_list.append( approx )
                active_vars.append( range(i+1) )
            tm_approx = MAPS.CommonBasisIntegratedExponentialTriangularTransportMap(
                active_vars,
                approx_list)
            tm_approx_list.append( tm_approx )
        self.tm_approx = MAPS.CompositeMap(tm_approx_list[0], tm_approx_list[1])
        self.params = {}
        self.params['params_t'] = None
        # self.coeffs = np.zeros(self.tm_approx.n_coeffs)
        self.coeffs = npr.randn(self.tm_approx.n_coeffs) / 10.

#
# LIST COMPOSITE MAP
#
class ListCompositeCommonBasisTotOrdIntegratedExponentialTTM( object ):
    def build_tm_approx(self):
        import SpectralToolbox.Spectral1D as S1D
        import TransportMaps as TM
        import TransportMaps.Functionals as FUNC
        import TransportMaps.Maps as MAPS
        tm_approx_list = []
        for i in range(3):
            # Build the transport map (isotropic for each entry)
            approx_list = []
            active_vars = []
            c_basis_list = [S1D.HermiteProbabilistsPolynomial() for i in range(self.dim)]
            e_basis_list = [S1D.ConstantExtendedHermiteProbabilistsFunction(self.order)
                            for i in range(self.dim)]
            for i in range(self.dim):
                c_orders_list = ([self.order] * i) + [0]
                c_approx = FUNC.LinearSpanApproximation(
                    c_basis_list[:i+1], spantype='total', order_list=c_orders_list)
                e_orders_list = [self.order] * (i+1)
                e_approx = FUNC.LinearSpanApproximation(
                    e_basis_list[:i+1], spantype='total', order_list=e_orders_list)
                approx = FUNC.MonotonicIntegratedExponentialApproximation(
                    c_approx, e_approx)
                approx_list.append( approx )
                active_vars.append( range(i+1) )
            tm_approx = MAPS.CommonBasisIntegratedExponentialTriangularTransportMap(
                active_vars,
                approx_list)
            tm_approx_list.append( tm_approx )
        self.tm_approx = MAPS.ListCompositeMap(tm_approx_list)
        self.params = {}
        self.params['params_t'] = None
        # self.coeffs = np.zeros(self.tm_approx.n_coeffs)
        self.coeffs = npr.randn(self.tm_approx.n_coeffs) / 10.

# LINEAR MAP
class OnTheFly_LinearTM_DerivativeChecks( LinearTM,
                                          OnTheFly_TM_DerivativeChecks):
    @unittest.skip("Not implemented")
    def test_grad_a(self):
        pass
    @unittest.skip("Not implemented")
    def test_hess_a(self):
        pass
    @unittest.skip("Not implemented")
    def test_action_hess_a(self):
        pass
    @unittest.skip("Not implemented")
    def test_tuple_grad_a(self):
        pass
    @unittest.skip("Not implemented")
    def test_grad_a_partial_xd(self):
        pass
    @unittest.skip("Not implemented")
    def test_grad_a_log_det_grad_x(self):
        pass
    @unittest.skip("Not implemented")
    def test_hess_a_log_det_grad_x(self):
        pass
    @unittest.skip("Not implemented")
    def test_action_hess_a_log_det_grad_x(self):
        pass
    @unittest.skip("Not implemented")
    def test_grad_a_inverse(self):
        pass
    @unittest.skip("Not implemented")
    def test_grad_a_grad_x(self):
        pass
    @unittest.skip("Not implemented")
    def test_grad_a_hess_x(self):
        pass
    @unittest.skip("Not implemented")
    def test_grad_a_hess_x_log_det_grad_x(self):
        pass
        
# FULL ORDER LINEAR SPAN
class OnTheFly_LinearSpanTM_DerivativeChecks( LinearSpanTTM,
                                              OnTheFly_TM_DerivativeChecks ):
    @unittest.skip("Expected to fail")
    def test_inverse(self):
        pass
    @unittest.skip("Expected to fail")
    def test_grad_a_inverse(self):
        pass
    @unittest.skip("Expected to fail")
    def test_grad_x_inverse(self):
        pass
    @unittest.skip("Expected to fail")
    def test_log_det_grad_x_inverse(self):
        pass

class PrecompUni_LinearSpanTM_DerivativeChecks( LinearSpanTTM,
                                                PrecompUni_TM_DerivativeChecks ):
    @unittest.skip("Expected to fail")
    def test_inverse(self):
        pass
    @unittest.skip("Expected to fail")
    def test_grad_a_inverse(self):
        pass
    @unittest.skip("Expected to fail")
    def test_grad_x_inverse(self):
        pass
    @unittest.skip("Expected to fail")
    def test_log_det_grad_x_inverse(self):
        pass

class PrecompMulti_LinearSpanTM_DerivativeChecks( LinearSpanTTM,
                                                  PrecompMulti_TM_DerivativeChecks ):
    @unittest.skip("Expected to fail")
    def test_inverse(self):
        pass
    @unittest.skip("Expected to fail")
    def test_grad_a_inverse(self):
        pass

class OnTheFly_CommonBasisLinearSpanTM_DerivativeChecks( CommonBasisLinearSpanTTM,
                                                         OnTheFly_TM_DerivativeChecks ):
    @unittest.skip("Expected to fail")
    def test_inverse(self):
        pass
    @unittest.skip("Expected to fail")
    def test_grad_a_inverse(self):
        pass
    @unittest.skip("Expected to fail")
    def test_grad_x_inverse(self):
        pass
    @unittest.skip("Expected to fail")
    def test_log_det_grad_x_inverse(self):
        pass

class PrecompUni_CommonBasisLinearSpanTM_DerivativeChecks( CommonBasisLinearSpanTTM,
                                                           PrecompUni_TM_DerivativeChecks ):
    @unittest.skip("Expected to fail")
    def test_inverse(self):
        pass
    @unittest.skip("Expected to fail")
    def test_grad_a_inverse(self):
        pass
    @unittest.skip("Expected to fail")
    def test_grad_x_inverse(self):
        pass
    @unittest.skip("Expected to fail")
    def test_log_det_grad_x_inverse(self):
        pass

# TOTAL ORDER LINEAR SPAN
class OnTheFly_TotOrdLinearSpanTM_DerivativeChecks( TotOrdLinearSpanTTM,
                                                    OnTheFly_TM_DerivativeChecks ):
    @unittest.skip("Expected to fail")
    def test_inverse(self):
        pass
    @unittest.skip("Expected to fail")
    def test_grad_a_inverse(self):
        pass
    @unittest.skip("Expected to fail")
    def test_grad_x_inverse(self):
        pass
    @unittest.skip("Expected to fail")
    def test_log_det_grad_x_inverse(self):
        pass

class PrecompUni_TotOrdLinearSpanTM_DerivativeChecks( TotOrdLinearSpanTTM,
                                                      PrecompUni_TM_DerivativeChecks ):
    @unittest.skip("Expected to fail")
    def test_inverse(self):
        pass
    @unittest.skip("Expected to fail")
    def test_grad_a_inverse(self):
        pass
    @unittest.skip("Expected to fail")
    def test_grad_x_inverse(self):
        pass
    @unittest.skip("Expected to fail")
    def test_log_det_grad_x_inverse(self):
        pass

class PrecompMulti_TotOrdLinearSpanTM_DerivativeChecks( TotOrdLinearSpanTTM,
                                                        PrecompMulti_TM_DerivativeChecks ):
    @unittest.skip("Expected to fail")
    def test_inverse(self):
        pass
    @unittest.skip("Expected to fail")
    def test_grad_a_inverse(self):
        pass
    @unittest.skip("Expected to fail")
    def test_grad_x_inverse(self):
        pass
    @unittest.skip("Expected to fail")
    def test_log_det_grad_x_inverse(self):
        pass

class OnTheFly_CommonBasisTotOrdLinearSpanTM_DerivativeChecks( CommonBasisTotOrdLinearSpanTTM,
                                                               OnTheFly_TM_DerivativeChecks ):
    @unittest.skip("Expected to fail")
    def test_inverse(self):
        pass
    @unittest.skip("Expected to fail")
    def test_grad_a_inverse(self):
        pass
    @unittest.skip("Expected to fail")
    def test_grad_x_inverse(self):
        pass
    @unittest.skip("Expected to fail")
    def test_log_det_grad_x_inverse(self):
        pass

class PrecompUni_CommonBasisTotOrdLinearSpanTM_DerivativeChecks( CommonBasisTotOrdLinearSpanTTM,
                                                                 PrecompUni_TM_DerivativeChecks ):
    @unittest.skip("Expected to fail")
    def test_inverse(self):
        pass
    @unittest.skip("Expected to fail")
    def test_grad_a_inverse(self):
        pass
    @unittest.skip("Expected to fail")
    def test_grad_x_inverse(self):
        pass
    @unittest.skip("Expected to fail")
    def test_log_det_grad_x_inverse(self):
        pass

# FULL ORDER INTEGRATED EXPONENTIAL
class OnTheFly_IntegratedExponentialTM_DerivativeChecks( IntegratedExponentialTTM,
                                                         OnTheFly_TM_DerivativeChecks ):
    pass

class PrecompUni_IntegratedExponentialTM_DerivativeChecks( IntegratedExponentialTTM,
                                                           PrecompUni_TM_DerivativeChecks ):
    pass

class PrecompMulti_IntegratedExponentialTM_DerivativeChecks( IntegratedExponentialTTM,
                                                             PrecompMulti_TM_DerivativeChecks ):
    pass

class OnTheFly_CommonBasisIntegratedExponentialTM_DerivativeChecks(
        CommonBasisIntegratedExponentialTTM,
        OnTheFly_TM_DerivativeChecks ):
    pass

class PrecompUni_CommonBasisIntegratedExponentialTM_DerivativeChecks(
        CommonBasisIntegratedExponentialTTM,
        PrecompUni_TM_DerivativeChecks ):
    pass

# TOTAL ORDER INTEGRATED EXPONENTIAL
class OnTheFly_TotOrdIntegratedExponentialTM_DerivativeChecks( TotOrdIntegratedExponentialTTM,
                                                               OnTheFly_TM_DerivativeChecks ):
    pass

class PrecompUni_TotOrdIntegratedExponentialTM_DerivativeChecks( TotOrdIntegratedExponentialTTM,
                                                                 PrecompUni_TM_DerivativeChecks ):
    pass

class PrecompMulti_TotOrdIntegratedExponentialTM_DerivativeChecks( TotOrdIntegratedExponentialTTM,
                                                                   PrecompMulti_TM_DerivativeChecks ):
    pass

class OnTheFly_CommonBasisTotOrdIntegratedExponentialTM_DerivativeChecks(
        CommonBasisTotOrdIntegratedExponentialTTM,
        OnTheFly_TM_DerivativeChecks ):
    pass

class PrecompUni_CommonBasisTotOrdIntegratedExponentialTM_DerivativeChecks(
        CommonBasisTotOrdIntegratedExponentialTTM,
        PrecompUni_TM_DerivativeChecks ):
    pass

# TOTAL ORDER INVERSE INTEGRATED EXPONENTIAL
class OnTheFly_InverseCommonBasisTotOrdIntegratedExponentialTM_DerivativeChecks(
        InverseCommonBasisTotOrdIntegratedExponentialTTM,
        OnTheFly_InverseTM_DerivativeChecks ):
    pass

# TOTAL ORDER COMPOSITE INTEGRATED EXPONENTIAL
class OnTheFly_CompositeCommonBasisTotOrdIntegratedExponentialTM_DerivativeChecks(
        CompositeCommonBasisTotOrdIntegratedExponentialTTM,
        OnTheFly_CompositeTM_DerivativeChecks ):
    pass

# TOTAL ORDER LIST COMPOSITE INTEGRATED EXPONENTIAL
class OnTheFly_ListCompositeCommonBasisTotOrdIntegratedExponentialTM_DerivativeChecks(
        ListCompositeCommonBasisTotOrdIntegratedExponentialTTM,
        OnTheFly_CompositeTM_DerivativeChecks ):
    @unittest.skip("Not Implemented")
    def test_hess_x_log_det_grad_x(self):
        pass
    @unittest.skip("Not Implemented")
    def test_action_hess_x_log_det_grad_x(self):
        pass

# FULL ORDER INTEGRATED SQUARED
class OnTheFly_IntegratedSquaredTM_DerivativeChecks( IntegratedSquaredTTM,
                                                         OnTheFly_TM_DerivativeChecks ):
    pass

class PrecompUni_IntegratedSquaredTM_DerivativeChecks( IntegratedSquaredTTM,
                                                           PrecompUni_TM_DerivativeChecks ):
    pass

class PrecompMulti_IntegratedSquaredTM_DerivativeChecks( IntegratedSquaredTTM,
                                                             PrecompMulti_TM_DerivativeChecks ):
    pass

class OnTheFly_CommonBasisIntegratedSquaredTM_DerivativeChecks(
        CommonBasisIntegratedSquaredTTM,
        OnTheFly_TM_DerivativeChecks ):
    pass

class PrecompUni_CommonBasisIntegratedSquaredTM_DerivativeChecks(
        CommonBasisIntegratedSquaredTTM,
        PrecompUni_TM_DerivativeChecks ):
    pass

# TOTAL ORDER INTEGRATED SQUARED
class OnTheFly_TotOrdIntegratedSquaredTM_DerivativeChecks( TotOrdIntegratedSquaredTTM,
                                                               OnTheFly_TM_DerivativeChecks ):
    pass

class PrecompUni_TotOrdIntegratedSquaredTM_DerivativeChecks( TotOrdIntegratedSquaredTTM,
                                                                 PrecompUni_TM_DerivativeChecks ):
    pass

class PrecompMulti_TotOrdIntegratedSquaredTM_DerivativeChecks( TotOrdIntegratedSquaredTTM,
                                                                   PrecompMulti_TM_DerivativeChecks ):
    pass

class OnTheFly_CommonBasisTotOrdIntegratedSquaredTM_DerivativeChecks(
        CommonBasisTotOrdIntegratedSquaredTTM,
        OnTheFly_TM_DerivativeChecks ):
    pass

class PrecompUni_CommonBasisTotOrdIntegratedSquaredTM_DerivativeChecks(
        CommonBasisTotOrdIntegratedSquaredTTM,
        PrecompUni_TM_DerivativeChecks ):
    pass
    
# TOTAL ORDER INVERSE INTEGRATED SQUARED
class OnTheFly_InverseCommonBasisTotOrdIntegratedSquaredTM_DerivativeChecks(
        InverseCommonBasisTotOrdIntegratedSquaredTTM,
        OnTheFly_InverseTM_DerivativeChecks ):
    pass

# FULL ORDER INTEGRATED SQUARED RBF
class OnTheFly_IntegratedSquaredRBFTM_DerivativeChecks( IntegratedSquaredRBFTTM,
                                                         OnTheFly_TM_DerivativeChecks ):
    pass

class PrecompUni_IntegratedSquaredRBFTM_DerivativeChecks( IntegratedSquaredRBFTTM,
                                                           PrecompUni_TM_DerivativeChecks ):
    pass

class PrecompMulti_IntegratedSquaredRBFTM_DerivativeChecks( IntegratedSquaredRBFTTM,
                                                             PrecompMulti_TM_DerivativeChecks ):
    pass

class OnTheFly_CommonBasisIntegratedSquaredRBFTM_DerivativeChecks(
        CommonBasisIntegratedSquaredRBFTTM,
        OnTheFly_TM_DerivativeChecks ):
    pass

class PrecompUni_CommonBasisIntegratedSquaredRBFTM_DerivativeChecks(
        CommonBasisIntegratedSquaredRBFTTM,
        PrecompUni_TM_DerivativeChecks ):
    pass

# TOTAL ORDER INTEGRATED SQUARED RBF
class OnTheFly_TotOrdIntegratedSquaredRBFTM_DerivativeChecks( TotOrdIntegratedSquaredRBFTTM,
                                                               OnTheFly_TM_DerivativeChecks ):
    pass

class PrecompUni_TotOrdIntegratedSquaredRBFTM_DerivativeChecks( TotOrdIntegratedSquaredRBFTTM,
                                                                 PrecompUni_TM_DerivativeChecks ):
    pass

class PrecompMulti_TotOrdIntegratedSquaredRBFTM_DerivativeChecks( TotOrdIntegratedSquaredRBFTTM,
                                                                   PrecompMulti_TM_DerivativeChecks ):
    pass

class OnTheFly_CommonBasisTotOrdIntegratedSquaredRBFTM_DerivativeChecks(
        CommonBasisTotOrdIntegratedSquaredRBFTTM,
        OnTheFly_TM_DerivativeChecks ):
    pass

class PrecompUni_CommonBasisTotOrdIntegratedSquaredRBFTM_DerivativeChecks(
        CommonBasisTotOrdIntegratedSquaredRBFTTM,
        PrecompUni_TM_DerivativeChecks ):
    pass

    
def build_suite(ttype='all'):
    # Linear map
    suite_of_ltm_dc = unittest.TestLoader().loadTestsFromTestCase(
        OnTheFly_LinearTM_DerivativeChecks )
    # Full order linear span
    suite_of_lstm_dc = unittest.TestLoader().loadTestsFromTestCase(
        OnTheFly_LinearSpanTM_DerivativeChecks )
    suite_pu_lstm_dc = unittest.TestLoader().loadTestsFromTestCase(
        PrecompUni_LinearSpanTM_DerivativeChecks )
    suite_pm_lstm_dc = unittest.TestLoader().loadTestsFromTestCase(
        PrecompMulti_LinearSpanTM_DerivativeChecks )
    suite_of_cblstm_dc = unittest.TestLoader().loadTestsFromTestCase(
        OnTheFly_CommonBasisLinearSpanTM_DerivativeChecks )
    suite_pu_cblstm_dc = unittest.TestLoader().loadTestsFromTestCase(
        PrecompUni_CommonBasisLinearSpanTM_DerivativeChecks )
    # Total order linear span
    suite_of_tolstm_dc = unittest.TestLoader().loadTestsFromTestCase(
        OnTheFly_TotOrdLinearSpanTM_DerivativeChecks )
    suite_pu_tolstm_dc = unittest.TestLoader().loadTestsFromTestCase(
        PrecompUni_TotOrdLinearSpanTM_DerivativeChecks )
    suite_pm_tolstm_dc = unittest.TestLoader().loadTestsFromTestCase(
        PrecompMulti_TotOrdLinearSpanTM_DerivativeChecks )
    suite_of_cbtolstm_dc = unittest.TestLoader().loadTestsFromTestCase(
        OnTheFly_CommonBasisTotOrdLinearSpanTM_DerivativeChecks )
    suite_pu_cbtolstm_dc = unittest.TestLoader().loadTestsFromTestCase(
        PrecompUni_CommonBasisTotOrdLinearSpanTM_DerivativeChecks )
    # Full order integrated exponential
    suite_of_ietm_dc = unittest.TestLoader().loadTestsFromTestCase(
        OnTheFly_IntegratedExponentialTM_DerivativeChecks )
    suite_pu_ietm_dc = unittest.TestLoader().loadTestsFromTestCase(
        PrecompUni_IntegratedExponentialTM_DerivativeChecks )
    suite_pm_ietm_dc = unittest.TestLoader().loadTestsFromTestCase(
        PrecompMulti_IntegratedExponentialTM_DerivativeChecks )
    suite_of_cbietm_dc = unittest.TestLoader().loadTestsFromTestCase(
        OnTheFly_CommonBasisIntegratedExponentialTM_DerivativeChecks )
    suite_pu_cbietm_dc = unittest.TestLoader().loadTestsFromTestCase(
        PrecompUni_CommonBasisIntegratedExponentialTM_DerivativeChecks )
    # Total order integrated exponential
    suite_of_toietm_dc = unittest.TestLoader().loadTestsFromTestCase(
        OnTheFly_TotOrdIntegratedExponentialTM_DerivativeChecks )
    suite_pu_toietm_dc = unittest.TestLoader().loadTestsFromTestCase(
        PrecompUni_TotOrdIntegratedExponentialTM_DerivativeChecks )
    suite_pm_toietm_dc = unittest.TestLoader().loadTestsFromTestCase(
        PrecompMulti_TotOrdIntegratedExponentialTM_DerivativeChecks )
    suite_of_cbtoietm_dc = unittest.TestLoader().loadTestsFromTestCase(
        OnTheFly_CommonBasisTotOrdIntegratedExponentialTM_DerivativeChecks )
    suite_pu_cbtoietm_dc = unittest.TestLoader().loadTestsFromTestCase(
        PrecompUni_CommonBasisTotOrdIntegratedExponentialTM_DerivativeChecks )
    # Inverse integrated exponential
    suite_of_icbtoietm_dc = unittest.TestLoader().loadTestsFromTestCase(
        OnTheFly_InverseCommonBasisTotOrdIntegratedExponentialTM_DerivativeChecks )
    # Composite integrated exponential
    suite_of_ccbtoietm_dc = unittest.TestLoader().loadTestsFromTestCase(
        OnTheFly_CompositeCommonBasisTotOrdIntegratedExponentialTM_DerivativeChecks )
    # List composite integrated exponential
    suite_of_lccbtoietm_dc = unittest.TestLoader().loadTestsFromTestCase(
        OnTheFly_ListCompositeCommonBasisTotOrdIntegratedExponentialTM_DerivativeChecks )
    # Full order integrated squared
    suite_of_istm_dc = unittest.TestLoader().loadTestsFromTestCase(
        OnTheFly_IntegratedSquaredTM_DerivativeChecks )
    suite_pu_istm_dc = unittest.TestLoader().loadTestsFromTestCase(
        PrecompUni_IntegratedSquaredTM_DerivativeChecks )
    suite_pm_istm_dc = unittest.TestLoader().loadTestsFromTestCase(
        PrecompMulti_IntegratedSquaredTM_DerivativeChecks )
    suite_of_cbistm_dc = unittest.TestLoader().loadTestsFromTestCase(
        OnTheFly_CommonBasisIntegratedSquaredTM_DerivativeChecks )
    suite_pu_cbistm_dc = unittest.TestLoader().loadTestsFromTestCase(
        PrecompUni_CommonBasisIntegratedSquaredTM_DerivativeChecks )
    # Total order integrated squared
    suite_of_toistm_dc = unittest.TestLoader().loadTestsFromTestCase(
        OnTheFly_TotOrdIntegratedSquaredTM_DerivativeChecks )
    suite_pu_toistm_dc = unittest.TestLoader().loadTestsFromTestCase(
        PrecompUni_TotOrdIntegratedSquaredTM_DerivativeChecks )
    suite_pm_toistm_dc = unittest.TestLoader().loadTestsFromTestCase(
        PrecompMulti_TotOrdIntegratedSquaredTM_DerivativeChecks )
    suite_of_cbtoistm_dc = unittest.TestLoader().loadTestsFromTestCase(
        OnTheFly_CommonBasisTotOrdIntegratedSquaredTM_DerivativeChecks )
    suite_pu_cbtoistm_dc = unittest.TestLoader().loadTestsFromTestCase(
        PrecompUni_CommonBasisTotOrdIntegratedSquaredTM_DerivativeChecks )
    # Inverse integrated squared
    suite_of_icbtoistm_dc = unittest.TestLoader().loadTestsFromTestCase(
        OnTheFly_InverseCommonBasisTotOrdIntegratedSquaredTM_DerivativeChecks )
    # Full order integrated squared rbf
    suite_of_isrbftm_dc = unittest.TestLoader().loadTestsFromTestCase(
        OnTheFly_IntegratedSquaredRBFTM_DerivativeChecks )
    suite_pu_isrbftm_dc = unittest.TestLoader().loadTestsFromTestCase(
        PrecompUni_IntegratedSquaredRBFTM_DerivativeChecks )
    suite_pm_isrbftm_dc = unittest.TestLoader().loadTestsFromTestCase(
        PrecompMulti_IntegratedSquaredRBFTM_DerivativeChecks )
    suite_of_cbisrbftm_dc = unittest.TestLoader().loadTestsFromTestCase(
        OnTheFly_CommonBasisIntegratedSquaredRBFTM_DerivativeChecks )
    suite_pu_cbisrbftm_dc = unittest.TestLoader().loadTestsFromTestCase(
        PrecompUni_CommonBasisIntegratedSquaredRBFTM_DerivativeChecks )
    # Total order integrated squared rbf
    suite_of_toisrbftm_dc = unittest.TestLoader().loadTestsFromTestCase(
        OnTheFly_TotOrdIntegratedSquaredRBFTM_DerivativeChecks )
    suite_pu_toisrbftm_dc = unittest.TestLoader().loadTestsFromTestCase(
        PrecompUni_TotOrdIntegratedSquaredRBFTM_DerivativeChecks )
    suite_pm_toisrbftm_dc = unittest.TestLoader().loadTestsFromTestCase(
        PrecompMulti_TotOrdIntegratedSquaredRBFTM_DerivativeChecks )
    suite_of_cbtoisrbftm_dc = unittest.TestLoader().loadTestsFromTestCase(
        OnTheFly_CommonBasisTotOrdIntegratedSquaredRBFTM_DerivativeChecks )
    suite_pu_cbtoisrbftm_dc = unittest.TestLoader().loadTestsFromTestCase(
        PrecompUni_CommonBasisTotOrdIntegratedSquaredRBFTM_DerivativeChecks )
    # GROUP SUITES
    suites_list = []
    if ttype in ['all','serial']:
        suites_list += [
            # Linear map
            suite_of_ltm_dc,
            
            # Linear span
            suite_of_lstm_dc, suite_pu_lstm_dc, suite_pm_lstm_dc,
            suite_of_cblstm_dc, suite_pu_cblstm_dc,
            suite_of_tolstm_dc, suite_pu_tolstm_dc, suite_pm_tolstm_dc,
            suite_of_cbtolstm_dc, suite_pu_cbtolstm_dc,

            # Integrated squared
            suite_of_istm_dc, suite_pu_istm_dc, suite_pm_istm_dc,
            # suite_of_cbistm_dc, suite_pu_cbistm_dc,
            suite_of_toistm_dc, suite_pu_toistm_dc, suite_pm_toistm_dc,
            # suite_of_cbtoistm_dc, suite_pu_cbtoistm_dc,
            # suite_of_icbtoistm_dc

            # Integrated squared rbf
            suite_of_isrbftm_dc, suite_pu_isrbftm_dc, suite_pm_isrbftm_dc,
            # suite_of_cbisrbftm_dc, suite_pu_cbisrbftm_dc,
            suite_of_toisrbftm_dc, suite_pu_toisrbftm_dc, suite_pm_toisrbftm_dc,
            # suite_of_cbtoisrbftm_dc, suite_pu_cbtoisrbftm_dc,

            # Integrated exponential
            suite_of_ietm_dc, suite_pu_ietm_dc, suite_pm_ietm_dc,
            suite_of_cbietm_dc, suite_pu_cbietm_dc,
            suite_of_toietm_dc, suite_pu_toietm_dc, suite_pm_toietm_dc,
            suite_of_cbtoietm_dc, suite_pu_cbtoietm_dc,
            suite_of_icbtoietm_dc,

            suite_of_ccbtoietm_dc,
            suite_of_lccbtoietm_dc,

        ]
    all_suites = unittest.TestSuite( suites_list )
    return all_suites

def run_tests(ttype='all'):
    all_suites = build_suite(ttype)
    # RUN
    unittest.TextTestRunner(verbosity=2).run(all_suites)

if __name__ == '__main__':
    run_tests()
